import pathlib
from collections import defaultdict
from functools import partial
from typing import Callable, Dict, List, Optional, Union

from robustnessgym.core.constants import CACHEDOPS
from robustnessgym.core.dataset import Batch, BatchOrDataset, Dataset
from robustnessgym.core.decorators import singlecolumn
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.operation import Operation
from robustnessgym.core.tools import (
    class_or_instancemethod,
    persistent_hash,
    recmerge,
    strings_as_json,
)


class CachedOperation(Operation):
    """Class to create CachedOperations."""

    # Path to a log directory
    logdir: pathlib.Path = pathlib.Path.home() / "robustnessgym/cachedops/"

    # Create a directory
    logdir.mkdir(parents=True, exist_ok=True)

    def __init__(
        self, apply_fn: Callable = None, identifier: Identifier = None, *args, **kwargs
    ):

        super(CachedOperation, self).__init__(
            apply_fn=apply_fn,
            identifiers=[identifier] if identifier else None,
            num_outputs=1,
            *args,
            **kwargs,
        )

    def __repr__(self):
        """Representation of a cached operation object.

        Returns: string representation
        """
        return str(self.identifier)

    @property
    def identifier(self):
        return self.identifiers[0]

    @staticmethod
    def store(batch: Batch, updates: List[Dict]) -> Batch:
        """Updates the cache of preprocessed information stored with each
        example in a batch.

        Args:
            batch: a batch of data
            updates: a list of dictionaries, one per example

        Returns: updated batch
        """
        if "cache" not in batch:
            batch["cache"] = [{} for _ in range(len(batch["index"]))]

        # For each example, recursively merge the example's original cache dictionary
        # with the update dictionary
        batch["cache"] = [
            recmerge(cache_dict, update_dict)
            for cache_dict, update_dict in zip(batch["cache"], updates)
        ]

        return batch

    @class_or_instancemethod
    def retrieve(
        self_or_cls,
        batch: Batch,
        columns: Union[List[str], List[List[str]]],
        proc_fns: Union[str, Callable, List[Union[str, Callable]]] = None,
        identifier: Union[str, Identifier] = None,
        reapply: bool = False,
        **kwargs,
    ) -> Optional[Union[Dict[tuple, List], List[List], Batch, List[Batch]]]:
        """Retrieve information from the cache.

        Args:
            batch: a batch of data
            columns: list of columns to retrieve cached information for
            proc_fns: list of processing functions to be executed left to right on
            the cached data
            identifier: name of the identifier to retrieve
            reapply: whether to recompute the cached operation at retrieval

        Returns: dict mapping a column to a list of length len(batch)
        """
        if not reapply:
            # Infer the most relevant key to retrieve if an identifier is not specified
            if not identifier:
                if isinstance(self_or_cls, type):
                    # cls
                    target_ident_key = self_or_cls.__name__
                else:
                    # self
                    target_ident_key = str(self_or_cls.identifier)

                cachedop_columns = defaultdict(list)
                for ident_key in batch.keys():
                    # Parse the identifier
                    ident = Identifier.parse(ident_key)
                    cachedop_columns[ident.without("columns")].append(ident)

                best_match, best_distance = None, 100000000
                for ident in cachedop_columns:
                    ident_key = str(ident)
                    # Pick the key that best matches the cls name or instance identifier
                    if (
                        ident_key.startswith(target_ident_key)
                        and len(ident_key.replace(target_ident_key, "")) < best_distance
                    ):
                        best_match = ident
                        best_distance = len(ident_key.replace(target_ident_key, ""))

                identifier = best_match

                # Still no identifier
                if not identifier:
                    raise ValueError(
                        f"Retrieval failed: couldn't find a key called "
                        f"{target_ident_key} in cache."
                    )

            try:
                if isinstance(columns, str) or (
                    isinstance(columns[0], str) and len(columns) == 1
                ):
                    if isinstance(columns, str):
                        columns = [columns]
                    # Retrieving single piece of information for a single column
                    retrieval = [
                        self_or_cls.decode(val)
                        for val in batch[identifier(columns=columns)]
                    ]

                elif isinstance(columns[0], str):
                    # Retrieving single piece of information `columns` list
                    retrieval = {
                        tuple(columns): [
                            self_or_cls.decode(val)
                            for val in batch[identifier(columns=columns)]
                        ]
                    }
                else:
                    # Retrieving multiple pieces of information
                    retrieval = {
                        tuple(cols_)
                        if len(cols_) > 1
                        else cols_[0]: [
                            self_or_cls.decode(val)
                            for val in batch[identifier(columns=cols_)]
                        ]
                        for cols_ in columns
                    }

            except KeyError:
                raise KeyError(
                    "Could not retrieve information for all columns. "
                    "If you're trying to retrieve information for multiple columns, "
                    "use columns=[[col_1], [col_2], ..] "
                    "instead of columns=[col_1, col_2, ..]."
                )

            # Check if the retrieved information needs to be processed
            if not proc_fns:
                return retrieval

            # Resolve the str proc_fns to callable(s)
            if isinstance(proc_fns, str):
                proc_fns = getattr(self_or_cls, proc_fns)
            elif isinstance(proc_fns, List):
                proc_fns = [
                    proc_fn
                    if isinstance(proc_fn, Callable)
                    else getattr(self_or_cls, proc_fn)
                    for proc_fn in proc_fns
                ]

            # Process and return the retrieved information
            if isinstance(proc_fns, Callable):
                if isinstance(retrieval, list):
                    return proc_fns(retrieval)
                else:
                    return {k: proc_fns(v) for k, v in retrieval.items()}

            if isinstance(retrieval, list):
                return [[proc_fn(v) for v in retrieval] for proc_fn in proc_fns]

            return [
                {k: proc_fn(v) for k, v in retrieval.items()} for proc_fn in proc_fns
            ]

        else:
            if proc_fns:
                print("Warning: proc_fns has no effect when reapply=True.")

            # Run the operation on the fly
            # TODO(karan): does this work for ops that require process_dataset
            if isinstance(columns, str) or (
                isinstance(columns[0], str) and len(columns) == 1
            ):
                if isinstance(columns, str):
                    columns = [columns]
                return (
                    self_or_cls(**kwargs).apply(batch=batch, columns=columns)
                    if isinstance(self_or_cls, type)
                    else self_or_cls.apply(batch=batch, columns=columns)
                )
            elif isinstance(columns[0], str):
                return {
                    tuple(columns): self_or_cls(**kwargs).apply(
                        batch=batch, columns=columns
                    )
                    if isinstance(self_or_cls, type)
                    else self_or_cls.apply(batch=batch, columns=columns)
                }
            return {
                tuple(cols_)
                if len(cols_) > 1
                else cols_[0]: self_or_cls(**kwargs).apply(batch=batch, columns=cols_)
                if isinstance(self_or_cls, type)
                else self_or_cls.apply(batch=batch, columns=cols_)
                for cols_ in columns
            }

    # @class_or_instancemethod
    # def retrieve(
    #         self_or_cls,
    #         batch: Batch,
    #         columns: Union[List[str], List[List[str]]],
    #         proc_fns: Union[str, Callable, List[Union[str, Callable]]] = None,
    #         identifier: Union[str, Identifier] = None,
    #         reapply: bool = False,
    #         **kwargs,
    # ) -> Optional[Union[Batch, List[Batch]]]:
    #     """Retrieve information from the cache.
    #
    #     Args:
    #         batch: a batch of data
    #         columns: list of columns to retrieve cached information for
    #         proc_fns: list of processing functions to be executed left to right on
    #         the cached data
    #         identifier: name of the identifier to retrieve
    #         reapply: whether to recompute the cached operation at retrieval
    #
    #     Returns: dict mapping a column to a list of length len(batch)
    #     """
    #     if not reapply:
    #         # Nothing to return if there's no cache
    #         if "cache" not in batch:
    #             raise ValueError(
    #                 "`cache` key missing: nothing has been cached yet. "
    #                 "Are you sure you ran a CachedOperation?"
    #             )
    #
    #         # Infer the most relevant key to retrieve if an identifier is not
    #         # specified
    #         if not identifier:
    #             if isinstance(self_or_cls, type):
    #                 # cls
    #                 target_ident_key = self_or_cls.__name__
    #             else:
    #                 # self
    #                 target_ident_key = str(self_or_cls.identifier)
    #
    #             best_match, best_distance = None, 100000000
    #             for ident_key in batch["cache"][0].keys():
    #                 # Pick the key that best matches the cls name or instance
    #                 # identifier
    #                 if (
    #                         ident_key.startswith(target_ident_key)
    #                         and len(
    #                     ident_key.replace(target_ident_key, "")) < best_distance
    #                 ):
    #                     best_match = ident_key
    #                     best_distance = len(ident_key.replace(target_ident_key, ""))
    #
    #             identifier = best_match
    #
    #             # Still no identifier
    #             if not identifier:
    #                 raise ValueError(
    #                     f"Retrieval failed: couldn't find a key called "
    #                     f"{target_ident_key} in cache."
    #                 )
    #
    #         try:
    #             if isinstance(columns[0], str):
    #                 retrieval = {
    #                     strings_as_json(columns): [
    #                         self_or_cls.decode(
    #                             cache[str(identifier)][strings_as_json(columns)]
    #                         )
    #                         for cache in batch["cache"]
    #                     ]
    #                 }
    #             else:
    #                 retrieval = {
    #                     strings_as_json(cols_): [
    #                         self_or_cls.decode(
    #                             cache[str(identifier)][strings_as_json(cols_)]
    #                         )
    #                         for cache in batch["cache"]
    #                     ]
    #                     for cols_ in columns
    #                 }
    #
    #         except KeyError:
    #             raise KeyError(
    #                 "Could not retrieve information for all columns. "
    #                 "If you're trying to retrieve information for multiple columns, "
    #                 "use columns=[[col_1], [col_2], ..] "
    #                 "instead of columns=[col_1, col_2, ..]."
    #             )
    #
    #         # Check if the retrieved information needs to be processed
    #         if not proc_fns:
    #             return retrieval
    #
    #         # Resolve the str proc_fns to callable(s)
    #         if isinstance(proc_fns, str):
    #             proc_fns = getattr(self_or_cls, proc_fns)
    #         elif isinstance(proc_fns, List):
    #             proc_fns = [
    #                 proc_fn
    #                 if isinstance(proc_fn, Callable)
    #                 else getattr(self_or_cls, proc_fn)
    #                 for proc_fn in proc_fns
    #             ]
    #
    #         # Process and return the retrieved information
    #         if isinstance(proc_fns, Callable):
    #             return {k: proc_fns(v) for k, v in retrieval.items()}
    #
    #         return [
    #             {k: proc_fn(v) for k, v in retrieval.items()} for proc_fn in proc_fns
    #         ]
    #
    #     else:
    #         if proc_fns:
    #             print("Warning: proc_fns has no effect when reapply=True.")
    #
    #         # Run the operation on the fly
    #         # TODO(karan): does this work for ops that require process_dataset
    #         if isinstance(columns[0], str):
    #             return {
    #                 strings_as_json(columns): self_or_cls(**kwargs).apply(
    #                     batch=batch, columns=columns
    #                 )
    #                 if isinstance(self_or_cls, type)
    #                 else self_or_cls.apply(batch=batch, columns=columns)
    #             }
    #         return {
    #             strings_as_json(cols_): self_or_cls(**kwargs).apply(
    #                 batch=batch, columns=cols_
    #             )
    #             if isinstance(self_or_cls, type)
    #             else self_or_cls.apply(batch=batch, columns=cols_)
    #             for cols_ in columns
    #         }

    @classmethod
    def available(cls, batch: Batch):
        """Check if the cached operation is available to retrieve in
        `batch`."""
        return any([key.startswith(cls.__name__) for key in batch.keys()])

    def get_cache_hash(self, columns: Optional[List[str]] = None):
        """Construct a hash that will be used to identify the application of a
        cached operation to the columns of a dataset."""

        val = hash(self)
        if columns:
            for key in columns:
                val ^= persistent_hash(key)
        return val

    def get_cache_file_name(self, columns=None):
        """Construct a file name for caching."""
        return "cache-" + str(abs(self.get_cache_hash(columns=columns))) + ".arrow"

    def prepare_batch(
        self,
        batch: Batch,
        columns: List[str],
        *args,
        **kwargs,
    ) -> Batch:
        """Preparation that is applied before the CachedOperation.

        This is provided as a convenience function that can be called by
        prepare_dataset.

        Args:
            batch: batch of examples
            columns: list of columns

        Returns: updated batch
        """
        raise NotImplementedError("Implement `prepare_batch`.")

    def prepare_dataset(
        self,
        dataset: Dataset,
        columns: List[str],
        batch_size: int = 32,
        *args,
        **kwargs,
    ) -> None:
        """Preparation that is applied before the CachedOperation.

        Many CachedOperations require a full pass over the dataset to precompute some
        variables before the core operation can actually be applied e.g. to create a
        Bag-of-Words representation, constructing a dataset vocabulary to keep only
        tokens that are frequently seen across the dataset.

        Args:
            dataset: Dataset
            columns: list of columns
            batch_size: batch size for .map(..)

        Returns: updated Dataset
        """

        # Set the data format
        dataset.set_format(columns)

        # Batch the dataset, and prepare each batch
        for batch in dataset.batch(batch_size):
            try:
                # Check if the `prepare_batch` function has been implemented
                self.prepare_batch(
                    batch=batch,
                    columns=columns,
                    *args,
                    **kwargs,
                )
            except NotImplementedError:
                break

        # Reset the data format
        dataset.reset_format()

        # # Apply preparation to the dataset
        # # TODO(karan): this is similar to the try except block for slicebuilders,
        # #  refactor
        # try:
        #     return dataset.map(
        #         partial(self.prepare_batch, columns=columns),
        #         batched=True,
        #         batch_size=batch_size,
        #         # The cache file name is a XOR of the interaction history and the
        #         # current operation
        #         # FIXME(karan): this is repeated
        #         cache_file_name=str(
        #             dataset.logdir
        #             / (
        #                 "cache-"
        #                 + str(
        #                     abs(
        #                         persistent_hash(str(dataset.identifier))
        #                         ^ dataset.hash_interactions()
        #                         ^ persistent_hash(
        #                             str(self.identifier) +
        #                             str(strings_as_json(columns))
        #                         )
        #                     )
        #                 )
        #                 + "-prep.arrow"
        #             )
        #         ),
        #     )
        # except:  # TypeError or PicklingError or AttributeError: # noqa
        #     # Batch the dataset, and process each batch
        #     all_batches = [
        #         self.prepare_batch(
        #             batch=batch,
        #             columns=columns,
        #         )
        #         for batch in dataset.batch(batch_size)
        #     ]
        #
        #     return dataset.map(
        #         lambda examples, indices: all_batches[indices[0] // batch_size],
        #         batched=True,
        #         batch_size=batch_size,
        #         with_indices=True,
        #         load_from_cache_file=False,
        #         # The cache file name is a XOR of the interaction history and the
        #         # current operation
        #         # FIXME(karan): this is repeated
        #         cache_file_name=str(
        #             dataset.logdir
        #             / (
        #                 "cache-"
        #                 + str(
        #                     abs(
        #                         persistent_hash(str(dataset.identifier))
        #                         ^ dataset.hash_interactions()
        #                         ^ persistent_hash(
        #                             str(self.identifier) +
        #                             str(strings_as_json(columns))
        #                         )
        #                     )
        #                 )
        #                 + "-prep.arrow"
        #             )
        #         ),
        #     )

    def apply(self, batch: Batch, columns: List[str], *args, **kwargs) -> List:
        """Implements the core functionality of the cached operation."""
        pass

    def _construct_updates(self, encoded_outputs: List[str], columns: List[str]):
        return [
            {str(self.identifier): {strings_as_json(columns): val}}
            for val in encoded_outputs
        ]

    def process_batch(
        self,
        batch: Batch,
        columns: List[str],
        *args,
        **kwargs,
    ) -> Batch:
        """Apply the cached operation to a batch."""
        assert len(set(columns) - set(batch.keys())) == 0, (
            f"All `columns` ({columns}) must be present in `batch` ("
            f"{list(batch.keys())})."
        )

        # Run the cached operation, and encode outputs (defaults to json.dumps)
        encoded_outputs = [
            self.encode(example_output)
            for example_output in self.apply(batch=batch, columns=columns)
        ]

        return {str(self.identifier(columns=columns)): encoded_outputs}

        # # Update the batch and return the updated batch
        # batch[str(self.identifier(columns=columns))] = encoded_outputs
        #
        # return batch

        # # Construct updates
        # updates = self._construct_updates(
        #     encoded_outputs=encoded_outputs, columns=columns
        # )
        #
        # # Update the cache and return the updated batch
        # return self.store(batch=batch, updates=updates)

    def process_dataset(
        self,
        dataset: Dataset,
        columns: List[str],
        batch_size: int = 32,
    ) -> Dataset:
        """Apply the cached operation to a dataset."""

        return dataset.update(
            partial(self.process_batch, columns=columns),
            batched=True,
            batch_size=batch_size,
        )

        # try:
        #     return dataset.map(
        #         partial(self.process_batch, columns=columns),
        #         batched=True,
        #         batch_size=batch_size,
        #         # The cache file name is a XOR of the interaction history and the
        #         # current operation
        #         cache_file_name=str(
        #             dataset.logdir
        #             / (
        #                 "cache-"
        #                 + str(
        #                     abs(
        #                         persistent_hash(str(dataset.identifier))
        #                         ^ dataset.hash_interactions()
        #                         ^ persistent_hash(
        #                             str(self.identifier) +
        #                             str(strings_as_json(columns))
        #                         )
        #                     )
        #                 )
        #                 + ".arrow"
        #             )
        #         ),
        #         # self.get_cache_file_name(columns=columns),
        #     )
        # except:  # noqa
        #     # Batch the dataset, and process each batch
        #     all_batches = [
        #         self.process_batch(
        #             batch=batch,
        #             columns=columns,
        #         )
        #         for batch in dataset.batch(batch_size)
        #     ]
        #
        #     return dataset.map(
        #         lambda examples, indices: all_batches[indices[0] // batch_size],
        #         batched=True,
        #         batch_size=batch_size,
        #         with_indices=True,
        #         load_from_cache_file=False,
        #         # The cache file name is a XOR of the interaction history and the
        #         # current operation
        #         cache_file_name=str(
        #             dataset.logdir
        #             / (
        #                 "cache-"
        #                 + str(
        #                     abs(
        #                         persistent_hash(str(dataset.identifier))
        #                         ^ dataset.hash_interactions()
        #                         ^ persistent_hash(
        #                             str(self.identifier) +
        #                             str(strings_as_json(columns))
        #                         )
        #                     )
        #                 )
        #                 + ".arrow"
        #             )
        #         ),
        #     )

    def __call__(
        self, batch_or_dataset: BatchOrDataset, columns: List[str], batch_size: int = 32
    ) -> BatchOrDataset:

        if isinstance(batch_or_dataset, Dataset):

            # Check the InteractionTape to see if the CachedOperation was applied
            if batch_or_dataset.check_tape(
                path=[CACHEDOPS],
                identifiers=self.identifier,
                columns=columns,
            ):
                return batch_or_dataset

            # Prepare to apply the CachedOperation to the dataset
            self.prepare_dataset(
                dataset=batch_or_dataset,
                columns=columns,
                batch_size=batch_size,
            )

            # Apply the CachedOperation to the dataset
            dataset = self.process_dataset(
                dataset=batch_or_dataset,
                columns=columns,
                batch_size=batch_size,
            )

            # Update the InteractionTape with the applied CachedOperation
            dataset.update_tape(
                path=[CACHEDOPS],
                identifiers=self.identifier,
                columns=columns,
            )

            return dataset

        elif isinstance(batch_or_dataset, Dict):
            # Apply the CachedOperation
            updates = self.process_batch(batch=batch_or_dataset, columns=columns)
            # Merge assuming that updates doesn't affect existing columns
            assert len(set(updates.keys()).intersection(batch_or_dataset.keys())) == 0
            return {**batch_or_dataset, **updates}
        else:
            raise NotImplementedError


class SingleColumnCachedOperation(CachedOperation):
    def __call__(
        self, batch_or_dataset: BatchOrDataset, columns: List[str], batch_size: int = 32
    ) -> BatchOrDataset:
        """Apply independently to each column.

        Args:
            batch_or_dataset:
            columns:

        Returns:
        """
        # Iterate over the columns and apply
        for column in columns:
            batch_or_dataset = super(SingleColumnCachedOperation, self).__call__(
                batch_or_dataset=batch_or_dataset,
                columns=[column],
                batch_size=batch_size,
            )

        return batch_or_dataset

    @singlecolumn
    def apply(self, batch: Batch, columns: List[str], *args, **kwargs) -> List:
        return self.single_column_apply(batch[columns[0]])

    def single_column_apply(self, column_batch: List, **kwargs) -> List:
        raise NotImplementedError("Must implement single_column_apply.")


class ScoreOperation(CachedOperation):
    def apply(
        self, batch: Batch, columns: List[str], *args, **kwargs
    ) -> List[Union[int, float]]:
        return super().apply(batch, columns, *args, **kwargs)


def stow(
    dataset: Dataset,
    cached_ops: Dict[CachedOperation, List[List[str]]],
    batch_size: int = 32,
    load_from_cache_file: bool = True,
):
    """Apply a list of cached operations in sequence."""

    # Check the InteractionTape to remove CachedOperations that have already been stowed
    for cached_op, list_of_columns in list(cached_ops.items()):
        indices_to_remove = []
        for i, columns in enumerate(list(list_of_columns)):
            if dataset.check_tape(
                path=[CACHEDOPS],
                identifiers=cached_op.identifier,
                columns=columns,
            ):
                # Remove the columns at index i
                indices_to_remove.append(i)

        # Remove the columns that are already cached
        for index in sorted(indices_to_remove, reverse=True):
            columns = cached_ops[cached_op].pop(index)
            print(f"skipped: {cached_op.identifier} -> {columns}", flush=True)

        # Check if list_of_columns is now empty
        if not cached_ops[cached_op]:
            # Remove the op entirely
            cached_ops.pop(cached_op)

    for cached_op, list_of_columns in cached_ops.items():
        for columns in list_of_columns:
            dataset = cached_op(dataset, columns=columns, batch_size=batch_size)

    # def _map_fn(batch: Batch):
    #     """
    #     Consolidate the application of the CachedOperations passed to stow into a
    #     single mappable function.
    #     """
    #     for cached_op, list_of_columns in cached_ops.items():
    #         for columns in list_of_columns:
    #             batch = cached_op(batch, columns=columns)
    #
    #     return batch
    #
    # # Compute the hash value
    # val = 0
    # for cached_op, list_of_columns in cached_ops.items():
    #     for columns in list_of_columns:
    #         val ^= cached_op.get_cache_hash(columns=columns)
    #
    # # Combine with the hash for the dataset on which the cached ops are applied
    # val ^= persistent_hash(
    #     # TODO(karan): move this to Dataset
    #     "-".join(
    #         "-".join(str(k) + "-" + str(v) for k, v in f.items()) for f in
    #         dataset._data_files
    #     )
    # )
    #
    # # Map the cached operations over the dataset
    # try:
    #     dataset = dataset.map(
    #         _map_fn,
    #         batched=True,
    #         batch_size=32,
    #         cache_file_name='cache-' + str(abs(val)) + '.arrow',
    #         load_from_cache_file=load_from_cache_file
    #     )
    # except TypeError:
    #     # Batch the dataset, and process each batch
    #     all_batches = [_map_fn(batch=batch) for batch in dataset.batch(batch_size)]
    #
    #     # Update the dataset efficiently by reusing all_batches
    #     dataset = dataset.map(
    #         lambda examples, indices: all_batches[indices[0] // batch_size],
    #         batched=True,
    #         batch_size=batch_size,
    #         with_indices=True,
    #     )

    # Update the Dataset history
    for cached_op, list_of_columns in cached_ops.items():
        for columns in list_of_columns:
            dataset.update_tape(
                path=[CACHEDOPS], identifiers=cached_op.identifier, columns=columns
            )

    return dataset



"""File containing abstract base class for datasets."""
import abc
from contextlib import contextmanager
from copy import copy
from types import SimpleNamespace
from typing import Callable, Dict, List, Mapping, Optional, Sequence, Union

import cytoolz as tz
import numpy as np
import pandas as pd
import pyarrow as pa
from datasets.arrow_dataset import DatasetInfoMixin

from robustnessgym.core.tools import recmerge

Example = Dict
Batch = Dict[str, List]


class AbstractDataset(
    abc.ABC,
    DatasetInfoMixin,
):
    """An abstract dataset class."""

    all_columns: Optional[List[str]]
    visible_columns: Optional[List[str]]
    visible_rows: Optional[np.ndarray]
    _data: Union[Dict[str, List], pd.DataFrame, pa.Table]

    def __init__(self, *args, **kwargs):
        super(AbstractDataset, self).__init__(*args, **kwargs)

    def __repr__(self):
        return f"RG{self.__class__.__name__}" f"(num_rows: {self.num_rows})"

    def __len__(self):
        # If only a subset of rows are visible
        if self.visible_rows is not None:
            return len(self.visible_rows)

        # If there are columns, len of any column
        if self.column_names:
            return len(self._data[self.column_names[0]])
        return 0

    @property
    def column_names(self):
        """Column names in the dataset."""
        return self.all_columns

    @property
    def columns(self):
        """Column names in the dataset."""
        return self.column_names

    @property
    def num_rows(self):
        """Number of rows in the dataset."""
        return len(self)

    @property
    def shape(self):
        """Shape of the dataset (num_rows, num_columns)."""
        return self.num_rows, len(self.columns)

    @classmethod
    def _assert_columns_all_equal_length(cls, batch: Batch):
        """Check that all columns have the same length so that the data is
        tabular."""
        assert cls._columns_all_equal_length(
            batch
        ), "All columns must have equal length."

    @classmethod
    def _columns_all_equal_length(cls, batch: Batch):
        """Check that all columns have the same length so that the data is
        tabular."""
        if len(set([len(v) for k, v in batch.items()])) == 1:
            return True
        return False

    def _check_columns_exist(self, columns: List[str]):
        """Check that every column in `columns` exists."""
        for col in columns:
            assert col in self.all_columns, f"{col} is not a valid column."

    @abc.abstractmethod
    def _set_features(self):
        """Set the features of the dataset."""
        raise NotImplementedError

    def _initialize_state(self):
        """Dataset state initialization."""
        # Show all columns by default
        self.visible_columns = copy(self.all_columns)

        # Show all rows by default
        self.visible_rows = None

        # Set the features
        self._set_features()

    @abc.abstractmethod
    def add_column(self, column: str, values: List):
        """Add a column to the dataset."""
        raise NotImplementedError

    def set_visible_rows(self, indices: Optional[Sequence]):
        """Set the visible rows in the dataset."""
        if indices is None:
            self.visible_rows = None
        else:
            if len(indices):
                assert min(indices) >= 0 and max(indices) < len(self), (
                    f"Ensure min index {min(indices)} >= 0 and "
                    f"max index {max(indices)} < {len(self)}."
                )
            if self.visible_rows is not None:
                self.visible_rows = self.visible_rows[np.array(indices, dtype=int)]
            else:
                self.visible_rows = np.array(indices, dtype=int)

    def reset_visible_rows(self):
        """Reset to make all rows visible."""
        self.visible_rows = None

    @contextmanager
    def format(self, columns: List[str] = None):
        """Context where only `columns` will be visible."""
        # Get the current format
        current_format = self.get_format()

        if columns:
            # View only `columns`
            self.set_format(columns)
        else:
            # Use all columns
            self.set_format(self.column_names)
        try:
            yield
        finally:
            # Reset the format back
            self.set_format(current_format)

    def get_format(self) -> List[str]:
        """Get the dataset format."""
        return self.visible_columns

    def set_format(self, columns: List[str]):
        """Set the dataset format.

        Only `columns` are visible after set_format is invoked.
        """
        # Check that the columns exist
        self._check_columns_exist(columns)

        # Set visible columns
        self.visible_columns = columns

    def reset_format(self):
        """Reset the dataset format.

        All columns are visible.
        """
        # All columns are visible
        self.visible_columns = self.all_columns

    def batch(self, batch_size: int = 32, drop_last_batch: bool = False):
        """Batch the dataset.

        Args:
            batch_size: integer batch size
            drop_last_batch: drop the last batch if its smaller than batch_size

        Returns:
            batches of data
        """
        for i in range(0, len(self), batch_size):
            if drop_last_batch and i + batch_size > len(self):
                continue
            yield self[i : i + batch_size]

    @classmethod
    def from_batch(cls, batch: Batch):
        """Create an AbstractDataset from a batch."""
        return cls(batch)

    @classmethod
    def from_batches(cls, batches: List[Batch]):
        """Create an AbstractDataset from a list of batches."""
        return cls.from_batch(
            tz.merge_with(tz.compose(list, tz.concat), *batches),
        )

    def _example_or_batch_to_batch(
        self, example_or_batch: Union[Example, Batch]
    ) -> Batch:

        # Check if example_or_batch is a batch
        is_batch = all(
            [isinstance(v, List) for v in example_or_batch.values()]
        ) and self._columns_all_equal_length(example_or_batch)

        # Convert to a batch if not
        if not is_batch:
            batch = {k: [v] for k, v in example_or_batch.items()}
        else:
            batch = example_or_batch

        return batch

    @classmethod
    def _merge_batch_and_output(cls, batch: Batch, output: Batch):
        """Merge an output during .map() into a batch."""
        combined = batch
        for k in output.keys():
            if k not in batch:
                combined[k] = output[k]
            else:
                if isinstance(batch[k][0], dict) and isinstance(output[k][0], dict):
                    combined[k] = [
                        recmerge(b_i, o_i) for b_i, o_i in zip(batch[k], output[k])
                    ]
                else:
                    combined[k] = output[k]
        return combined

    @classmethod
    def _mask_batch(cls, batch: Batch, boolean_mask: List[bool]):
        """Remove elements in `batch` that are masked by `boolean_mask`."""
        return {
            k: [e for i, e in enumerate(v) if boolean_mask[i]] for k, v in batch.items()
        }

    def _inspect_function(
        self,
        function: Callable,
        with_indices: bool = False,
        batched: bool = False,
    ) -> SimpleNamespace:

        # Initialize variables to track
        no_output = dict_output = bool_output = list_output = False

        # If dict_output = True and `function` is used for updating the dataset
        # useful to know if any existing column is modified
        updates_existing_column = True
        existing_columns_updated = []

        # Run the function to test it
        if batched:
            if with_indices:
                output = function(self[:2], range(2))
            else:
                output = function(self[:2])

        else:
            if with_indices:
                output = function(self[0], 0)
            else:
                output = function(self[0])

        if isinstance(output, Mapping):
            # `function` returns a dict output
            dict_output = True

            # Set of columns that are updated
            existing_columns_updated = set(self.all_columns).intersection(
                set(output.keys())
            )

            # Check if `function` updates an existing column
            if len(existing_columns_updated) == 0:
                updates_existing_column = False

        elif output is None:
            # `function` returns None
            no_output = True
        elif isinstance(output, bool):
            # `function` returns a bool
            bool_output = True
        elif isinstance(output, list):
            # `function` returns a list
            list_output = True
            if batched and isinstance(output[0], bool):
                # `function` returns a bool per example
                bool_output = True

        return SimpleNamespace(
            dict_output=dict_output,
            no_output=no_output,
            bool_output=bool_output,
            list_output=list_output,
            updates_existing_column=updates_existing_column,
            existing_columns_updated=existing_columns_updated,
        )

    @abc.abstractmethod
    def append(
        self,
        example_or_batch: Union[Example, Batch],
    ) -> None:
        """Append a batch of data to the dataset.

        `batch` must have the same columns as the dataset (regardless of
        what columns are visible).
        """
        raise NotImplementedError



# from typing import Callable, Dict, List, Tuple
#
# import numpy as np
#
# from robustnessgym.core.identifier import Identifier
# from robustnessgym.slicebuilders.subpopulation import Subpopulation
#
#
# class Position:
#     def __init__(
#         self,
#         identifiers: List[Identifier] = None,
#         position_fn: Callable = None,
#         *args,
#         **kwargs
#     ):
#
#         if not identifiers:
#             identifiers = [
#                 Identifier(
#                     _name=self.__class__.__name__,
#                     gte=interval[0],
#                     lte=interval[1],
#                     score_fn=score_fn,
#                 )
#                 for interval in intervals
#             ]
#
#         # Set the intervals
#         self.intervals = intervals
#         self.left_limits = np.array([interval[0] for interval in intervals])
#         self.right_limits = np.array([interval[1] for interval in intervals])
#
#         # Assign the score fn
#         if score_fn:
#             self.score = score_fn
#
#     def score(
#         self, batch: Dict[str, List], keys: List[str], *args, **kwargs
#     ) -> np.ndarray:
#         raise NotImplementedError("Return a vector of float scores for each example.")
#
#     def apply(
#         self,
#         slice_membership: np.ndarray,
#         batch: Dict[str, List],
#         keys: List[str],
#         *args,
#         **kwargs
#     ) -> np.ndarray:
#         # Keep track of the score of each example
#         scores = self.score(batch=batch, keys=keys)
#
#         assert scores.shape == (
#             slice_membership.shape[0],
#         ), "Must have exactly one score per example."
#
#         return (
#             (self.left_limits <= scores[:, np.newaxis])
#             & (scores[:, np.newaxis] <= self.right_limits)
#         ).astype(np.int32)



import hashlib
import inspect
import json
from collections import defaultdict
from functools import partial
from typing import Callable, Dict, List, Mapping, Optional, Sequence

import cytoolz as tz
import progressbar
import yaml


def convert_to_batch_fn(function: Callable, with_indices: bool):
    """Batch a function that applies to an example."""

    def _function(
        batch: Dict[str, List], indices: Optional[List[int]], *args, **kwargs
    ):
        # Pull out the batch size
        batch_size = len(batch[list(batch.keys())[0]])

        # Iterate and apply the function
        outputs = None
        for i in range(batch_size):

            # Apply the unbatched function
            if with_indices:
                output = function(
                    {k: v[i] for k, v in batch.items()},
                    indices[i],
                    *args,
                    **kwargs,
                )
            else:
                output = function(
                    {k: v[i] for k, v in batch.items()},
                    *args,
                    **kwargs,
                )

            if i == 0:
                # Create an empty dict or list for the outputs
                outputs = defaultdict(list) if isinstance(output, dict) else []

            # Append the output
            if isinstance(output, dict):
                for k in output.keys():
                    outputs[k].append(output[k])
            else:
                outputs.append(output)

        if isinstance(outputs, dict):
            return dict(outputs)
        return outputs

    if with_indices:
        # Just return the function as is
        return _function
    else:
        # Wrap in a lambda to apply the indices argument
        return lambda batch, *args, **kwargs: _function(batch, None, *args, **kwargs)


def recmerge(*objs, merge_sequences=False):
    """Recursively merge an arbitrary number of collections. For conflicting
    values, later collections to the right are given priority. By default
    (merge_sequences=False), sequences are treated as a normal value and not
    merged.

    Args:
        *objs: collections to merge
        merge_sequences: whether to merge values that are sequences

    Returns: merged collection
    """
    if isinstance(objs, tuple) and len(objs) == 1:
        # A squeeze operation since merge_with generates tuple(list_of_objs,)
        objs = objs[0]
    if all([isinstance(obj, Mapping) for obj in objs]):
        # Merges all the collections, recursively applies merging to the combined values
        return tz.merge_with(partial(recmerge, merge_sequences=merge_sequences), *objs)
    elif all([isinstance(obj, Sequence) for obj in objs]) and merge_sequences:
        # Merges sequence values by concatenation
        return list(tz.concat(objs))
    else:
        # If colls does not contain mappings, simply pick the last one
        return tz.last(objs)


def persistent_hash(s: str):
    """Compute a hash that persists across multiple Python sessions for a
    string."""
    return int(hashlib.sha224(s.encode()).hexdigest(), 16)


def strings_as_json(strings: List[str]):
    """Convert a list of strings to JSON.

    Args:
        strings: A list of str.

    Returns: JSON dump of the strings.
    """
    if len(strings) > 1:
        return json.dumps(strings)
    elif len(strings) == 1:
        return strings[0]
    else:
        return ""


def get_default_args(func) -> dict:
    """Inspect a function to get arguments that have default values.

    Args:
        func: a Python function

    Returns: dictionary where keys correspond to arguments, and values correspond to
    their defaults.
    """
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


class DownloadProgressBar:
    def __init__(self):
        self.pbar = None

    def __call__(self, block_num, block_size, total_size):
        if not self.pbar:
            self.pbar = progressbar.ProgressBar(
                maxval=total_size if total_size > 0 else 1e-2
            )
            self.pbar.start()

        downloaded = block_num * block_size
        if downloaded < total_size:
            self.pbar.update(downloaded)
        else:
            self.pbar.finish()


def prettyprint(s: str) -> None:
    """Prettyprint with YAML.

    Args:
        s: string
    """
    if hasattr(s, "__dict__"):
        print(yaml.dump(s.__dict__))
    elif isinstance(s, dict):
        print(yaml.dump(s))
    else:
        print(s)


def get_all_leaf_paths(coll):
    """Returns a list of paths to all leaf nodes in a nested dict.

    Paths can travel through lists and the index is inserted into the
    path.
    """
    if isinstance(coll, Mapping):
        return list(
            tz.concat(
                map(
                    lambda t: list(map(lambda p: [t[0]] + p, get_all_leaf_paths(t[1]))),
                    coll.items(),
                )
            )
        )

    elif isinstance(coll, list):
        return list(
            tz.concat(
                map(
                    lambda t: list(map(lambda p: [t[0]] + p, get_all_leaf_paths(t[1]))),
                    enumerate(coll),
                )
            )
        )
    else:
        return [[]]


def get_all_paths(coll, prefix_path=(), stop_at=None, stop_below=None):
    """Given a collection, by default returns paths to all the leaf nodes.

    Use stop_at to truncate paths at the given key. Use stop_below to
    truncate paths one level below the given key.
    """
    assert (
        stop_at is None or stop_below is None
    ), "Only one of stop_at or stop_below can be used."
    if stop_below is not None and stop_below in str(
        tz.last(tz.take(len(prefix_path) - 1, prefix_path))
    ):
        return [[]]
    if stop_at is not None and stop_at in str(tz.last(prefix_path)):
        return [[]]
    if isinstance(coll, Mapping) or isinstance(coll, list):
        if isinstance(coll, Mapping):
            items = coll.items()
        else:
            items = enumerate(coll)

        return list(
            tz.concat(
                map(
                    lambda t: list(
                        map(
                            lambda p: [t[0]] + p,
                            get_all_paths(
                                t[1],
                                prefix_path=list(prefix_path) + [t[0]],
                                stop_at=stop_at,
                                stop_below=stop_below,
                            ),
                        )
                    ),
                    items,
                )
            )
        )
    else:
        return [[]]


def get_only_paths(coll, pred, prefix_path=(), stop_at=None, stop_below=None):
    """Get all paths that satisfy the predicate fn pred.

    First gets all paths and then filters them based on pred.
    """
    all_paths = get_all_paths(
        coll, prefix_path=prefix_path, stop_at=stop_at, stop_below=stop_below
    )
    return list(filter(pred, all_paths))


class class_or_instancemethod(classmethod):
    def __get__(self, instance, type_):
        descr_get = super().__get__ if instance is None else self.__func__.__get__
        return descr_get(instance, type_)



from __future__ import annotations

import json
from typing import List, Union

import cytoolz as tz

from robustnessgym.core.constants import (
    ATTACK,
    CACHEDOPS,
    SLICEBUILDERS,
    SUBPOPULATION,
    TRANSFORMATION,
)
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.tools import persistent_hash, strings_as_json


class InteractionTape:
    """Class for interaction tape to keep track of applied ops."""

    def __init__(self):
        # Keep track of the history
        self.history = {}

    def __repr__(self):
        return f"{self.__class__.__name__}(interactions={len(self.history)})"

    def __hash__(self):
        val = 0
        for (identifier, json_columns) in self.history:
            val ^= persistent_hash(str(identifier) + str(json_columns))
        return val

    def dumps(self):
        return json.dumps(
            {
                json.dumps((identifier.dumps(), json_columns)): idx
                for (identifier, json_columns), idx in self.history.items()
            }
        )

    @classmethod
    def loads(cls, s: str):
        tape = InteractionTape()
        history = json.loads(s)
        history = {
            tuple(json.loads(json_tuple)): idx for json_tuple, idx in history.items()
        }
        tape.history = {
            (Identifier.loads(identifier), json_columns): idx
            for (identifier, json_columns), idx in history.items()
        }

        return tape

    def update(self, identifier: Union[str, Identifier], columns: List[str]) -> None:
        """Update the interaction tape with information about an interaction.

        Args:
            identifier: Identifier for the interaction used.
            columns: list of columns on which the interaction was applied.

        Returns: True if the interaction was added to the tape, False if it was
        already applied before.
        """
        if isinstance(identifier, str):
            identifier = Identifier(_name=identifier)
        elif isinstance(identifier, Identifier):
            pass
        else:
            raise ValueError(
                f"Parameter `identifier` should be an instance of class Identifier "
                f"or str, "
                f"not {type(identifier)}."
            )

        # Dump the column names to JSON
        json_columns = strings_as_json(strings=columns)

        # Check if the entry is not in the history
        if (identifier, json_columns) not in self.history:
            # Give it the next index
            self.history[(identifier, json_columns)] = len(self.history)

    def check(self, identifier: Union[str, Identifier], columns: List[str]) -> bool:
        """

        Args:
            identifier:
            columns:

        Returns:

        """
        if not (isinstance(identifier, str) or isinstance(identifier, Identifier)):
            raise ValueError(
                f"Parameter `identifier` should be an instance of class Identifier "
                f"or str, "
                f"not {type(identifier)}."
            )

        # Dump the column names to JSON
        json_columns = strings_as_json(strings=columns)

        # Check if the entry is already in the history
        if (identifier, json_columns) in self.history:
            return True
        return False


class InteractionTapeHierarchyMixin:
    def __init__(self):
        self.interactions = {
            CACHEDOPS: InteractionTape(),
            SLICEBUILDERS: {
                SUBPOPULATION: InteractionTape(),
                TRANSFORMATION: InteractionTape(),
                ATTACK: InteractionTape(),
            },
        }

    def hash_interactions(self):
        v = 0
        for path in [
            [CACHEDOPS],
            [SLICEBUILDERS, SUBPOPULATION],
            [SLICEBUILDERS, TRANSFORMATION],
            [SLICEBUILDERS, ATTACK],
        ]:
            v ^= self.fetch_tape(path=path).__hash__()
        return v

    def dumps_interactions(self):
        return json.dumps(
            {
                CACHEDOPS: self.interactions[CACHEDOPS].dumps(),
                SLICEBUILDERS: {
                    SUBPOPULATION: self.interactions[SLICEBUILDERS][
                        SUBPOPULATION
                    ].dumps(),
                    TRANSFORMATION: self.interactions[SLICEBUILDERS][
                        TRANSFORMATION
                    ].dumps(),
                    ATTACK: self.interactions[SLICEBUILDERS][ATTACK].dumps(),
                },
            }
        )

    @classmethod
    def loads_interactions(cls, s: str) -> InteractionTapeHierarchyMixin:
        tape_hierarchy = InteractionTapeHierarchyMixin()
        interactions = json.loads(s)
        tape_hierarchy.interactions = {
            CACHEDOPS: InteractionTape.loads(interactions[CACHEDOPS]),
            SLICEBUILDERS: {
                SUBPOPULATION: InteractionTape.loads(
                    interactions[SLICEBUILDERS][SUBPOPULATION]
                ),
                TRANSFORMATION: InteractionTape.loads(
                    interactions[SLICEBUILDERS][TRANSFORMATION]
                ),
                ATTACK: InteractionTape.loads(interactions[SLICEBUILDERS][ATTACK]),
            },
        }
        return tape_hierarchy

    def update_tape(
        self,
        path: List[str],
        identifiers: Union[Identifier, List[Identifier]],
        columns: List[str],
    ):
        """Update the tape.

        Args:
            path: Location of the InteractionTape in the hierarchy.
            identifiers:
            columns:

        Returns:
        """
        # Fetch the tape
        tape = self.fetch_tape(path=path)

        # Update it
        if isinstance(identifiers, Identifier) or isinstance(identifiers, str):
            return tape.update(identifier=identifiers, columns=columns)
        else:
            return [
                tape.update(identifier=identifier, columns=columns)
                for identifier in identifiers
            ]

    def check_tape(
        self,
        path: List[str],
        identifiers: Union[Identifier, List[Identifier]],
        columns: List[str],
    ):
        """Check the tape.

        Args:

            path:
            identifiers:
            columns:

        Returns:
        """
        # Fetch the tape
        tape = self.fetch_tape(path=path)

        # Check it
        if isinstance(identifiers, Identifier) or isinstance(identifiers, str):
            return tape.check(identifier=identifiers, columns=columns)
        else:
            return [
                tape.check(identifier=identifier, columns=columns)
                for identifier in identifiers
            ]

    def fetch_tape(self, path: List[str]) -> InteractionTape:
        """Fetch an InteractionTape."""
        return tz.get_in(path, self.interactions)



from functools import partial
from typing import Dict, List, Sequence, Tuple

import numpy as np
from numpy.linalg import norm
from scipy.linalg import svd
from scipy.stats import spearmanr

from robustnessgym.cachedops.similarity import RougeMatrix, RougeScore
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.tools import strings_as_json
from robustnessgym.slicebuilders.subpopulations.score import ScoreSubpopulation


class RougeScoreSubpopulation(ScoreSubpopulation):
    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        metric: Sequence[str] = ("rouge1", "fmeasure"),
        *args,
        **kwargs,
    ):
        super(RougeScoreSubpopulation, self).__init__(
            intervals=intervals,
            identifiers=[
                Identifier(
                    _name=self.__class__.__name__,
                    gte=interval[0],
                    lte=interval[1],
                    metric=metric,
                )
                for interval in intervals
            ],
            *args,
            **kwargs,
        )

        # Assign the metric
        self.metric = metric

    def score(
        self, batch: Dict[str, List], columns: List[str], *args, **kwargs
    ) -> np.ndarray:
        assert len(columns) == 2, "Must have exactly 2 columns."

        # Retrieve Rouge scores
        scores = RougeScore.retrieve(
            batch=batch,
            columns=columns,
            proc_fns=partial(RougeScore.select, metric=self.metric),
        )[strings_as_json(columns)]

        return np.array(scores)


class Abstractiveness(RougeScoreSubpopulation):
    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        metric: Sequence[str] = ("rouge1", "precision"),
    ):
        super(Abstractiveness, self).__init__(
            intervals=intervals,
            metric=metric,
        )


class Distillation(RougeScoreSubpopulation):
    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        metric: Sequence[str] = ("rouge1", "recall"),
    ):
        super(Distillation, self).__init__(
            intervals=intervals,
            metric=metric,
        )


class RougeMatrixScoreSubpopulation(ScoreSubpopulation):
    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        metric: Sequence[str] = ("rouge1", "fmeasure"),
        *args,
        **kwargs,
    ):
        assert (
            len(metric) == 2
        ), "Must pass in both rouge score and one of precision/recall/fmeasure."
        super(RougeMatrixScoreSubpopulation, self).__init__(
            intervals=intervals,
            identifiers=[
                Identifier(
                    _name=self.__class__.__name__,
                    gte=interval[0],
                    lte=interval[1],
                    metric=metric,
                )
                for interval in intervals
            ],
            *args,
            **kwargs,
        )

        # Assign the metric
        self.metric = metric

    def reduce(self, matrices: List[np.ndarray]) -> np.ndarray:
        raise NotImplementedError

    def score(
        self, batch: Dict[str, List], columns: List[str], *args, **kwargs
    ) -> np.ndarray:
        assert len(columns) == 2, "Must have exactly 2 columns."

        # Retrieve the relevant Rouge matrices
        matrices = RougeMatrix.retrieve(
            batch=batch,
            columns=columns,
            proc_fns=partial(RougeMatrix.select, metric=self.metric),
        )[strings_as_json(columns)]

        return self.reduce(matrices)


class Position(RougeMatrixScoreSubpopulation):
    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        metric: Sequence[str] = ("rouge1", "fmeasure"),
    ):
        super(Position, self).__init__(
            intervals=intervals,
            metric=metric,
        )

    def reduce(self, matrices: List[np.ndarray]) -> np.ndarray:
        # Compute position of best-matched sentence in source document
        # Then compute mean position, capturing where position mostly comes from
        return np.array(
            [np.mean(np.argmax(mat, axis=0)) / mat.shape[0] for mat in matrices]
        )


class Dispersion(RougeMatrixScoreSubpopulation):
    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        metric: Sequence[str] = ("rouge1", "fmeasure"),
    ):
        super(Dispersion, self).__init__(
            intervals=intervals,
            metric=metric,
        )

    def reduce(self, matrices: List[np.ndarray]) -> np.ndarray:
        # Compute position of best-matched sentence in source document
        # Then compute std dev of position, capturing how spread out the positions are
        return np.array(
            [np.std(np.argmax(mat, axis=0) / mat.shape[0]) for mat in matrices]
        )


class Ordering(RougeMatrixScoreSubpopulation):
    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        metric: Sequence[str] = ("rouge1", "fmeasure"),
    ):
        super(Ordering, self).__init__(
            intervals=intervals,
            metric=metric,
        )

    def reduce(self, matrices: List[np.ndarray]) -> np.ndarray:
        # Compute position of best-matched sentence in source document
        # Then compute spearman correlation of position with range(..),
        # capturing whether the order of information is reversed
        return np.array(
            [
                spearmanr(
                    np.arange(mat.shape[1]) / mat.shape[0],
                    np.argmax(mat, axis=0) / mat.shape[0],
                )[0]
                for mat in matrices
            ]
        )


class NuclearNorm(RougeMatrixScoreSubpopulation):
    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        metric: Sequence[str] = ("rouge1", "fmeasure"),
    ):
        super(NuclearNorm, self).__init__(
            intervals=intervals,
            metric=metric,
        )

    def reduce(self, matrices: List[np.ndarray]) -> np.ndarray:
        return np.array(
            [
                np.sum(np.abs(svd(mat, full_matrices=False, compute_uv=False)))
                for mat in matrices
            ]
        )


class SpectralNorm(RougeMatrixScoreSubpopulation):
    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        metric: Sequence[str] = ("rouge1", "fmeasure"),
    ):
        super(SpectralNorm, self).__init__(
            intervals=intervals,
            metric=metric,
        )

    def reduce(self, matrices: List[np.ndarray]) -> np.ndarray:
        return np.array(
            [
                np.max(np.abs(svd(mat, full_matrices=False, compute_uv=False)))
                for mat in matrices
            ]
        )


class FrobeniusNorm(RougeMatrixScoreSubpopulation):
    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        metric: Sequence[str] = ("rouge1", "fmeasure"),
    ):
        super(FrobeniusNorm, self).__init__(
            intervals=intervals,
            metric=metric,
        )

    def reduce(self, matrices: List[np.ndarray]) -> np.ndarray:
        return np.array([norm(mat) for mat in matrices])



from __future__ import annotations

from collections import Counter
from typing import Dict, List, Tuple

import numpy as np

from robustnessgym.cachedops.spacy import Spacy
from robustnessgym.core.dataset import Dataset
from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.subpopulation import Subpopulation


# TODO(karan): Inherit from MultiScoreSubpopulation
class EntityFrequency(Subpopulation, Spacy):
    def __init__(self, entity_thresholds: List[Tuple[str, List[int]]], *args, **kwargs):

        identifiers = []
        for entity_type, thresholds in entity_thresholds:
            for threshold in thresholds:
                identifiers.append(
                    Identifier(
                        _name=self.__class__.__name__,
                        entity_type=entity_type,
                        threshold=threshold,
                    )
                )

        super(EntityFrequency, self).__init__(identifiers, *args, **kwargs)

        if len(entity_thresholds) == 0:
            raise ValueError("At least one entity type required")

        for entity_type, _ in entity_thresholds:
            if entity_type not in [
                "PERSON",
                "NORP",
                "FAC",
                "ORG",
                "GPE",
                "LOC",
                "PRODUCT",
                "EVENT",
                "WORK_OF_ART",
                "LAW",
                "LANGUAGE",
                "DATE",
                "TIME",
                "PERCENT",
                "MONEY",
                "QUANTITY",
                "ORDINAL",
                "CARDINAL",
            ]:
                raise ValueError(f"Invalid entity type: {entity_type}")

        # List of tuples, each of which contains an entity type and a list of
        # associated frequency thresholds
        self.entity_thresholds = entity_thresholds

    @classmethod
    def from_dataset(
        cls, dataset: Dataset, entity_percentiles: List[Tuple[str, List[float]]]
    ) -> EntityFrequency:
        """Determine thresholds from dataset and specific percentiles."""
        raise NotImplementedError

    def apply(
        self,
        slice_membership: np.ndarray,
        batch: Dict[str, List],
        columns: List[str],
        *args,
        **kwargs,
    ) -> np.ndarray:

        if len(columns) != 1:
            raise ValueError("Only one key allowed")
        key = columns[0]

        for i, cache_item in enumerate(batch["cache"]):
            entities = cache_item["Spacy"][key]["ents"]
            entity_types = [ent["label"] for ent in entities]
            counts = Counter(entity_types)
            slice_ndx = 0
            for entity_type, thresholds in self.entity_thresholds:
                for threshold in thresholds:
                    if counts[entity_type] >= threshold:
                        slice_membership[i, slice_ndx] = 1
                    slice_ndx += 1

        return slice_membership



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['robustnessgym',
 'robustnessgym.active',
 'robustnessgym.cachedops',
 'robustnessgym.cachedops.allen',
 'robustnessgym.core',
 'robustnessgym.core.dataformats',
 'robustnessgym.logging',
 'robustnessgym.slicebuilders',
 'robustnessgym.slicebuilders.attacks',
 'robustnessgym.slicebuilders.subpopulations',
 'robustnessgym.slicebuilders.transformations',
 'robustnessgym.tasks']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.29.21,<0.30.0',
 'cytoolz>=0.11.0,<0.12.0',
 'datasets>=1.1.3,<2.0.0',
 'dill>=0.3.3,<0.4.0',
 'fastBPE>=0.1.0,<0.2.0',
 'fuzzywuzzy>=0.18.0,<0.19.0',
 'ipywidgets>=7.6.2,<8.0.0',
 'jsonlines>=1.2.0,<2.0.0',
 'jupyterlab>=3.0.0,<4.0.0',
 'kaleido==0.1.0',
 'multiprocess>=0.70.11,<0.71.0',
 'numpy>=1.18.0,<2.0.0',
 'omegaconf>=2.0.5,<3.0.0',
 'plotly>=4.14.1,<5.0.0',
 'progressbar>=2.5,<3.0',
 'pyahocorasick>=1.4.0,<2.0.0',
 'python-Levenshtein>=0.12.0,<0.13.0',
 'pytorch-lightning>=1.1.2,<2.0.0',
 'scikit-learn>=0.24.1,<0.25.0',
 'semver>=2.13.0,<3.0.0',
 'tqdm>=4.27.0,<5.0.0',
 'transformers>=4.0.0,<5.0.0']

extras_require = \
{'adversarial': ['textattack>=0.2.15,<0.3.0'],
 'augmentation': ['nlpaug>=1.1.1,<2.0.0'],
 'ludwig': ['tensorflow>=2.3.0,<3.0.0', 'ludwig>=0.3,<0.4'],
 'summarization': ['rouge-score>=0.0.4,<0.0.5'],
 'text': ['nltk>=3.5,<4.0',
          'textblob>=0.15.3,<0.16.0',
          'spacy>=2.3.5,<3.0.0',
          'allennlp>=1.3.0,<2.0.0',
          'allennlp-models>=1.3.0,<2.0.0',
          'stanza>=1.1.1,<2.0.0'],
 'vision': ['torchvision>=0.8.0,<0.9.0']}

setup_kwargs = {
    'name': 'robustnessgym',
    'version': '0.0.4a2',
    'description': 'Robustness Gym is an evaluation toolkit for natural language processing.',
    'long_description': '\n\n<img src="docs/logo.png" width="50" height="50" alt="Robustness Gym logo"/> Robustness Gym\n================================\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/robustness-gym/robustness-gym/CI)\n![GitHub](https://img.shields.io/github/license/robustness-gym/robustness-gym)\n[![codecov](https://codecov.io/gh/robustness-gym/robustness-gym/branch/main/graph/badge.svg?token=MOLQYUSYQU)](https://codecov.io/gh/robustness-gym/robustness-gym)\n[![Documentation Status](https://readthedocs.org/projects/robustnessgym/badge/?version=latest)](https://robustnessgym.readthedocs.io/en/latest/?badge=latest)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![website](https://img.shields.io/badge/website-live-brightgreen)](https://robustnessgym.com)\n\nRobustness Gym is a Python evaluation toolkit for natural language processing. \n\n#### Coming Soon\n- Examples & tutorials\n- More documentation\n- Contributing guidelines\n\n\n### About\nRobustness Gym is being developed to address challenges in evaluating machine\n learning models today. You can read more about the ideas underlying Robustness Gym\n  in our paper on [arXiv](https://arxiv.org/pdf/2101.04840.pdf). We also have a\n   [website](https://robustnessgym.com).\n\nThe Robustness Gym project is an ongoing collaboration between [Stanford Hazy\n Research](https://hazyresearch.stanford.edu), [Salesforce Research](https://einstein.ai\n ) and [UNC Chapel-Hill](http://murgelab.cs.unc.edu/). \n\n_Note: Robustness Gym is in alpha, so expect frequent updates in the coming weeks and \nmonths. Reach out to kgoel [at] cs [dot] stanford [dot] edu if you\'d like to become an active contributor, \nor if you work on an interesting NLP task that you\'d like to see supported. \nFeel free to raise issues on GitHub for bugs/feature requests._\n\n### Installation\n```\npip install robustnessgym\n```\n\n### Robustness Gym in 5 minutes\n\n#### Datasets that extend Huggingface `datasets`\n```python\n# robustnessgym.Dataset wraps datasets.Dataset\nfrom robustnessgym import Dataset\n\n# Use Dataset.load_dataset(..) exactly like datasets.load_dataset(..) \ndataset = Dataset.load_dataset(\'boolq\')\ndataset = Dataset.load_dataset(\'boolq\', split=\'train[:10]\')\n```\n\n#### Cache information\n```python\n# Get a dataset\nfrom robustnessgym import Dataset\ndataset = Dataset.load_dataset(\'boolq\')\n\n# Run the Spacy pipeline\nfrom robustnessgym import Spacy\nspacy = Spacy()\n# .. on the \'question\' column of the dataset\ndataset = spacy(batch_or_dataset=dataset, \n                columns=[\'question\'])\n\n\n# Run the Stanza pipeline\nfrom robustnessgym import Stanza\nstanza = Stanza()\n# .. on both the question and passage columns of a batch\ndataset = stanza(batch_or_dataset=dataset[:32], \n                 columns=[\'question\', \'passage\'])\n\n# .. use any of the other built-in operations in Robustness Gym!\n\n\n# Or, create your own CachedOperation\nfrom robustnessgym import CachedOperation, Identifier\nfrom robustnessgym.core.decorators import singlecolumn\n\n# Write a silly function that operates on a single column of a batch\n@singlecolumn\ndef silly_fn(batch, columns):\n    """\n    Capitalize text in the specified column of the batch.\n    """\n    column_name = columns[0]\n    assert type(batch[column_name]) == str, "Must apply to text column."\n    return [text.capitalize() for text in batch[column_name]] \n\n# Wrap the silly function in a CachedOperation\nsilly_op = CachedOperation(apply_fn=silly_fn,\n                           identifier=Identifier(_name=\'SillyOp\'))\n\n# Apply it to a dataset\ndataset = silly_op(batch_or_dataset=dataset, \n                   columns=[\'question\'])\n```\n\n\n#### Retrieve cached information\n```python\nfrom robustnessgym import Spacy, Stanza, CachedOperation\n\n# Take a batch of data\nbatch = dataset[:32]\n\n# Retrieve the (cached) results of the Spacy CachedOperation \nspacy_information = Spacy.retrieve(batch, columns=[\'question\'])\n\n# Retrieve the tokens returned by the Spacy CachedOperation\ntokens = Spacy.retrieve(batch, columns=[\'question\'], proc_fns=Spacy.tokens)\n\n# Retrieve the entities found by the Stanza CachedOperation\nentities = Stanza.retrieve(batch, columns=[\'passage\'], proc_fns=Stanza.entities)\n\n# Retrieve the capitalized output of the silly_op\ncapitalizations = CachedOperation.retrieve(batch,\n                                           columns=[\'question\'],\n                                           identifier=silly_op.identifier)\n\n# Retrieve it directly using the silly_op\ncapitalizations = silly_op.retrieve(batch, columns=[\'question\'])\n\n# Retrieve the capitalized output and lower-case it during retrieval\ncapitalizations = silly_op.retrieve(\n    batch,\n    columns=[\'question\'],\n    proc_fns=lambda decoded_batch: [x.lower() for x in decoded_batch]\n)\n```\n\n#### Create subpopulations\n```python\nfrom robustnessgym import Spacy, ScoreSubpopulation\n\ndef length(batch, columns):\n    """\n    Length using cached Spacy tokenization.\n    """\n    column_name = columns[0]\n    # Take advantage of previously cached Spacy informations\n    tokens = Spacy.retrieve(batch, columns, proc_fns=Spacy.tokens)[column_name]\n    return [len(tokens_) for tokens_ in tokens]\n\n# Create a subpopulation that buckets examples based on length\nlength_subpopulation = ScoreSubpopulation(intervals=[(0, 10), (10, 20)],\n                                          score_fn=length)\n\ndataset, slices, membership = length_subpopulation(dataset, columns=[\'question\'])\n# dataset is updated with slice information\n# slices is a list of 2 Slice objects\n# membership is a matrix of shape (n x 2)\n```\n',
    'author': 'Robustness Gym',
    'author_email': 'kgoel@cs.stanford.edu',
    'maintainer': 'Karan Goel',
    'maintainer_email': 'kgoel@cs.stanford.edu',
    'url': 'https://robustnessgym.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List

import cytoolz as tz
import torch

from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.transformation import SingleColumnTransformation

try:
    import fastBPE  # noqa
except ImportError:
    _fastbpe_available = False
else:
    _fastbpe_available = True


# TODO(karan): spec requirements (fastBPE)
class FairseqBacktranslation(SingleColumnTransformation):
    """Class for performing backtranslation using torchhub fairseq models."""

    def __init__(
        self,
        n_src2tgt: int = 1,
        n_tgt2src: int = 1,
        langs: str = "en2de",
        torchhub_dir: str = None,
        device: str = "cuda",
        src2tgt_topk: int = 1000,
        src2tgt_temp: float = 1.0,
        tgt2src_topk: int = 1000,
        tgt2src_temp: float = 1.0,
    ):

        if not _fastbpe_available:
            raise ImportError(
                "fastBPE not available for import. Please install fastBPE with pip "
                "install fastBPE."
            )

        super(FairseqBacktranslation, self).__init__(
            identifiers=Identifier.range(
                n=n_src2tgt * n_tgt2src,
                _name=self.__class__.__name__,
                langs=langs,
                src2tgt_topk=src2tgt_topk,
                src2tgt_temp=src2tgt_temp,
                tgt2src_topk=tgt2src_topk,
                tgt2src_temp=tgt2src_temp,
            )
        )

        # Set the parameters
        self.n_src2tgt = n_src2tgt
        self.n_tgt2src = n_tgt2src
        self.src2tgt_topk = src2tgt_topk
        self.src2tgt_temp = src2tgt_temp
        self.tgt2src_topk = tgt2src_topk
        self.tgt2src_temp = tgt2src_temp

        # Setup the backtranslation models
        self.src2tgt, self.tgt2src = self.load_models(
            langs=langs,
            torchhub_dir=torchhub_dir,
            # self.logdir if not torchhub_dir else torchhub_dir,
            device=device,
        )

    @staticmethod
    def load_models(
        langs: str,
        torchhub_dir: str = None,
        device: str = "cuda",
        half_precision: bool = False,
    ):
        if torchhub_dir:
            # Set the directory where the models will be stored.
            torch.hub.set_dir(torchhub_dir)

        if langs == "en2de":
            # Round-trip translations between English and German
            src2tgt = torch.hub.load(
                "pytorch/fairseq",
                "transformer.wmt19.en-de.single_model",
                tokenizer="moses",
                bpe="fastbpe",
            )

            tgt2src = torch.hub.load(
                "pytorch/fairseq",
                "transformer.wmt19.de-en.single_model",
                tokenizer="moses",
                bpe="fastbpe",
            )

        elif langs == "en2ru":
            # Round-trip translations between English and Russian
            src2tgt = torch.hub.load(
                "pytorch/fairseq",
                "transformer.wmt19.en-ru.single_model",
                tokenizer="moses",
                bpe="fastbpe",
            )

            tgt2src = torch.hub.load(
                "pytorch/fairseq",
                "transformer.wmt19.ru-en.single_model",
                tokenizer="moses",
                bpe="fastbpe",
            )
        else:
            raise NotImplementedError

        # Convert to half precision
        if half_precision:
            return src2tgt.to(device).half(), tgt2src.to(device).half()
        return src2tgt.to(device), tgt2src.to(device)

    def single_column_apply(self, column_batch: List) -> List[List]:
        """Perform backtranslation using the fairseq pretrained translation
        models."""
        # Encode the source sentences
        src_sentences = column_batch
        src_sentences_bin = [self.src2tgt.encode(e)[:1024] for e in src_sentences]

        # Translate it
        tgt_sentences = self.src2tgt.generate(
            src_sentences_bin,
            beam=self.n_src2tgt,
            sampling=True,
            sampling_topk=self.src2tgt_topk,
            temperature=self.src2tgt_temp,
            skip_invalid_size_inputs=True,
        )

        # Back-translate: moving tokens to CPU because of an error otherwise
        src_paraphrases = self.tgt2src.generate(
            [e["tokens"].cpu() for l in tgt_sentences for e in l],
            beam=self.n_tgt2src,
            sampling=True,
            sampling_topk=self.tgt2src_topk,
            temperature=self.tgt2src_temp,
            skip_invalid_size_inputs=True,
        )

        # Flatten out all the translations into one giant list
        flat_src_paraphrases = list(
            tz.concat(
                map(
                    lambda l: list(map(lambda e: self.tgt2src.decode(e["tokens"]), l)),
                    src_paraphrases,
                )
            )
        )

        # Partition so that we get n_src2tgt * n_tgt2src paraphrases per input sentence
        return list(
            tz.partition_all(
                len(flat_src_paraphrases) // len(src_sentences), flat_src_paraphrases
            )
        )



from __future__ import annotations

from collections import OrderedDict
from typing import Callable, Collection, Dict, List, Tuple

from datasets.features import FeatureType

from robustnessgym.core.dataset import Dataset
from robustnessgym.core.tools import get_all_paths


class Schema:
    """Class for task input/output schemas in Robustness Gym."""

    def __init__(
        self,
        features: OrderedDict,
        grounding_candidates: Dict[str, Collection],
    ):
        # Store the features and grounding candidates
        self.features = features
        self.grounding_candidates = grounding_candidates
        self.reversed_grounding_candidates = {
            v: k for k, values in self.grounding_candidates.items() for v in values
        }

    @classmethod
    def from_columns(
        cls,
        features: Dict[str, FeatureType],
        columns: List[str],
    ) -> Schema:
        """Create a schema using features and columns."""
        for col in columns:
            assert col in features, f"Column {col} must be in `features`."

        return Schema(
            features=OrderedDict({k: v for k, v in features.items() if k in columns}),
            grounding_candidates={k: {k} for k in columns},
        )

    @classmethod
    def for_dataset(
        cls,
        dataset: Dataset,
        input_columns: List[str],
        output_columns: List[str],
    ) -> Tuple[Schema, Schema]:
        """Create input and output schemas using features, input and output
        columns."""
        # Set the features
        features = dataset.features

        return Schema.from_columns(features, input_columns), Schema.from_columns(
            features, output_columns
        )

    def ground(self, features: Dict[str, FeatureType]):
        """

        Args:
            features: given by Dataset.features

        Returns: (grounding, reversed_grounding)

        """
        # For features, get the path to the leaves in the (potentially nested)
        # features dictionary
        flat_columns = get_all_paths(features)
        flat_columns = {
            tuple(path) if len(path) > 1 else path[0] for path in flat_columns
        }

        # Figure out the (reversed) grounding: map columns in the dataset to keys in
        # the schema
        reversed_grounding = {}
        for k in self.reversed_grounding_candidates:
            if ((isinstance(k, tuple) or isinstance(k, str)) and k in flat_columns) or (
                isinstance(k, Callable)
            ):
                reversed_grounding[k] = self.reversed_grounding_candidates[k]

        # Construct the grounding by reversing
        grounding = {v: k for k, v in reversed_grounding.items()}

        # Assert that the grounding covers the entire schema
        assert len(self.features) == len(grounding), "Grounding failed."

        # Assert that the grounded schema has the right types
        # FIXME(karan): Value == ClassLabel should be allowed: shouldn't break this
        # TODO(karan): if not, add code to automatically rejig the dataset in map_fn
        # for key in self.features:
        #     if isinstance(grounding[key], str):
        #         assert self.features[key] == features[grounding[key]]
        #     elif isinstance(grounding[key], tuple):
        #         assert self.features[key] == tz.get_in(grounding[key], features)

        return grounding, reversed_grounding

    def __repr__(self):
        features = "\n\t".join([f"{k}: {v}" for k, v in self.features.items()])
        return f"Schema(\n\t{features}\n)"

    def __len__(self):
        return len(self.features)

    @property
    def columns(self):
        """List of columns that participate in the schema."""
        return list(self.features.keys())



from .attack import Attack  # noqa
from .slicebuilder import SliceBuilder  # noqa
from .subpopulation import Subpopulation, SubpopulationCollection  # noqa
from .transformation import Transformation  # noqa



from typing import List

from robustnessgym.core.cachedops import SingleColumnCachedOperation

try:
    import stanza
except ImportError:
    _stanza_available = False
else:
    _stanza_available = True


class Stanza(SingleColumnCachedOperation):
    """Class for running the Stanza pipeline using a CachedOperation.

    URL: https://stanfordnlp.github.io/stanza/
    """

    def __init__(self):
        if not _stanza_available:
            raise ImportError(
                "Stanza not available for import. Install using " "\npip install stanza"
            )
        super(Stanza, self).__init__()

        self._download()
        self.nlp = stanza.Pipeline()

    def _download(self):
        stanza.download()

    @classmethod
    def encode(cls, obj: "stanza.Document") -> str:
        # Dump the Stanza Document to a string
        return obj.to_serialized()

    @classmethod
    def decode(cls, s: str):
        # Load the Stanza Document from the string
        return stanza.Document.from_serialized(s)

    def single_column_apply(self, column_batch: List, *args, **kwargs) -> List:
        # Create a doc for each example
        return [self.nlp(text) for text in column_batch]

    @classmethod
    def _get_attribute(
        cls, decoded_batch: List["stanza.Document"], attribute: str
    ) -> List:
        """Get an arbitrary attribute using doc.get(attribute) from a list of
        Stanza Documents."""
        return [doc.get(attribute) for doc in decoded_batch]

    @classmethod
    def lemma(cls, decoded_batch: List["stanza.Document"]) -> List[List[str]]:
        """For each example, returns the list of lemmatized words."""
        return cls._get_attribute(decoded_batch, "lemma")

    @classmethod
    def text(cls, decoded_batch: List["stanza.Document"]) -> List[List[str]]:
        """For each example, returns the list of tokens."""
        return cls._get_attribute(decoded_batch, "text")

    @classmethod
    def upos(cls, decoded_batch: List["stanza.Document"]) -> List[List[str]]:
        """For each example, returns the list of upos."""
        return cls._get_attribute(decoded_batch, "upos")

    @classmethod
    def xpos(cls, decoded_batch: List["stanza.Document"]) -> List[List[str]]:
        """For each example, returns the list of xpos."""
        return cls._get_attribute(decoded_batch, "xpos")

    @classmethod
    def feats(cls, decoded_batch: List["stanza.Document"]) -> List[List[str]]:
        """For each example, returns the list of feats."""
        return cls._get_attribute(decoded_batch, "feats")

    @classmethod
    def head(cls, decoded_batch: List["stanza.Document"]) -> List[List[str]]:
        """For each example, returns the list of head."""
        return cls._get_attribute(decoded_batch, "head")

    @classmethod
    def deprel(cls, decoded_batch: List["stanza.Document"]) -> List[List[str]]:
        """For each example, returns the list of deprel."""
        return cls._get_attribute(decoded_batch, "deprel")

    @classmethod
    def misc(cls, decoded_batch: List["stanza.Document"]) -> List[List[str]]:
        """For each example, returns the list of misc."""
        return cls._get_attribute(decoded_batch, "misc")

    @classmethod
    def entities(cls, decoded_batch: List["stanza.Document"]) -> List[List[str]]:
        """For each example, returns the list of entities."""
        return [doc.entities for doc in decoded_batch]

    @classmethod
    def id(cls, decoded_batch: List["stanza.Document"]) -> List[List[str]]:
        """For each example, returns the list of ids."""
        return cls._get_attribute(decoded_batch, "id")



import json
from typing import Dict, List, Optional, Sequence, Tuple, Union

import cytoolz as tz
import numpy as np
from multiprocess.pool import Pool
from tqdm import tqdm

from robustnessgym.core.constants import SLICEBUILDERS, SUBPOPULATION
from robustnessgym.core.dataset import Batch, Dataset
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.slice import Slice
from robustnessgym.core.tools import strings_as_json
from robustnessgym.slicebuilders.slicebuilder import SliceBuilder


class Subpopulation(SliceBuilder):
    def __init__(self, identifiers: List[Identifier], apply_fn=None, *args, **kwargs):
        super(Subpopulation, self).__init__(
            category=SUBPOPULATION,
            identifiers=identifiers,
            apply_fn=apply_fn,
            *args,
            **kwargs,
        )

    def apply(
        self,
        slice_membership: np.ndarray,
        batch: Batch,
        columns: List[str],
        *args,
        **kwargs,
    ) -> np.ndarray:
        raise NotImplementedError

    def process_batch(
        self,
        batch: Dict[str, List],
        columns: List[str],
        # mask: List[int] = None,
        *args,
        **kwargs,
    ) -> Tuple[Optional[List[Batch]], Optional[np.ndarray]]:

        # Determine the size of the batch
        batch_size = len(batch[list(batch.keys())[0]])

        # Construct the matrix of slice labels: (batch_size x num_slices)
        slice_membership = np.zeros((batch_size, self.num_slices), dtype=np.int32)

        # Apply the SliceBuilder's core functionality
        slice_membership = self.apply(slice_membership, batch, columns, *args, **kwargs)

        # # Store these slice labels
        # # TODO(karan): figure out how to set the alias
        # updates = self.construct_updates(
        #     slice_membership=slice_membership,
        #     columns=columns,
        #     mask=mask,
        #     compress=store_compressed,
        # )
        #
        # if store:
        #     batch = self.store(
        #         batch=batch,
        #         updates=updates,
        #     )

        return (
            None,  # self.filter_batch_by_slice_membership(batch, slice_membership),
            slice_membership,
        )

    def process_dataset(
        self,
        dataset: Dataset,
        columns: List[str],
        batch_size: int = 32,
        # mask: List[int] = None,
        num_proc: int = None,
        *args,
        **kwargs,
    ) -> Tuple[List[Slice], np.ndarray]:

        # Create slices using the dataset
        slices = [Slice(dataset) for _ in range(len(self.identifiers))]
        all_slice_memberships = []
        # Batch the dataset, and process each batch
        for batch in dataset.batch(batch_size):
            # Process the batch
            _, slice_memberships = self.process_batch(
                batch=batch,
                columns=columns,
                *args,
                **kwargs,
            )

            # Keep track of the slice memberships
            all_slice_memberships.append(slice_memberships)

        # Create a single slice label matrix
        slice_membership = np.concatenate(all_slice_memberships, axis=0)

        for i, sl in enumerate(slices):
            # Set the visible rows for each slice
            sl.set_visible_rows(np.where(slice_membership[:, i])[0])

            # Set the Slice category using the SliceBuilder's category
            sl.category = self.category

            # Append the the lineage
            sl.add_to_lineage(
                category=str(self.category.capitalize()),
                identifier=self.identifiers[i],
                columns=strings_as_json(columns),
            )

            # # Create the lineage
            # sl.lineage = [
            #     (str(Dataset.__name__), dataset.identifier),
            #     (
            #         str(self.category.capitalize()),
            #         self.identifiers[i],
            #         strings_as_json(columns),
            #     ),
            # ]
            # if isinstance(dataset, Slice):
            #     # Prepend the Slice's lineage instead, if the dataset was a slice
            #     sl.lineage = dataset.lineage + [
            #         (
            #             str(self.category.capitalize()),
            #             self.identifiers[i],
            #             strings_as_json(columns),
            #         )
            #     ]

        return slices, slice_membership

    def construct_updates(
        self,
        slice_membership: np.ndarray,
        columns: List[str],
        mask: List[int] = None,
        compress: bool = True,
    ):

        # Mask out components
        # TODO(karan): masking inside apply, but only if the components are computed
        #  independently

        # Construct a list of update dicts that contains the slice membership for
        # each example
        if compress:
            # TODO(karan): this will overwrite a previous application of the same
            #  Slicer right now, need a merge operation
            # Merge is just an append to whatever list already exists
            return [
                {
                    self.category: {
                        self.__class__.__name__: {
                            json.dumps(columns) if len(columns) > 1 else columns[0]: row
                        }
                    }
                }
                for row in (
                    slice_membership[:, np.logical_not(np.array(mask, dtype=bool))]
                    if mask
                    else slice_membership
                ).tolist()
            ]

        return [
            {
                self.category: {
                    self.__class__.__name__: {
                        str(self.identifiers[i]): {
                            json.dumps(columns)
                            if len(columns) > 1
                            else columns[0]: membership
                        }
                        for i, membership in enumerate(row)
                        if not mask or not mask[i]
                    },
                }
            }
            for row in slice_membership.tolist()
        ]

    @classmethod
    def union(
        cls, *slicemakers: SliceBuilder, identifier: Identifier = None
    ) -> SliceBuilder:
        """Combine a list of slicers using a union."""
        # Group the slicers based on their class
        grouped_slicers = tz.groupby(lambda s: s.__class__, slicemakers)

        # Join the slicers corresponding to each class, and flatten
        slicemakers = list(
            tz.concat(
                tz.itemmap(
                    lambda item: (item[0], item[0].join(*item[1])), grouped_slicers
                ).values()
            )
        )

        def apply_fn(slice_membership, batch, columns, *args, **kwargs):
            # Determine the size of the batch
            batch_size = len(batch[list(batch.keys())[0]])

            # Keep track of all the slice labels
            all_slice_membership = []

            # Run each slicemaker on the batch
            for slicemaker in slicemakers:
                all_slice_membership.append(
                    slicemaker.apply(
                        slice_membership=np.zeros(
                            (batch_size, slicemaker.num_slices), dtype=np.int32
                        ),
                        batch=batch,
                        columns=columns,
                    )
                )

            # Concatenate all the slice labels
            slice_membership = np.concatenate(all_slice_membership, axis=1)

            # Take the union over the slices (columns)
            slice_membership = np.any(slice_membership, axis=1).astype(np.int32)[
                :, np.newaxis
            ]

            return slice_membership

        return Subpopulation(identifiers=[identifier], apply_fn=apply_fn)

    @classmethod
    def intersection(
        cls, *slicemakers: SliceBuilder, identifier: Identifier = None
    ) -> SliceBuilder:
        """Combine a list of slicemakers using an intersection."""
        # Group the slicemakers based on their class
        grouped_slicemakers = tz.groupby(lambda s: s.__class__, slicemakers)

        # Join the slicemakers corresponding to each class, and flatten
        slicemakers = list(
            tz.concat(
                tz.itemmap(
                    lambda item: (item[0], item[0].join(*item[1])), grouped_slicemakers
                ).values()
            )
        )

        def apply_fn(slice_membership, batch, columns, *args, **kwargs):
            # Determine the size of the batch
            batch_size = len(batch[list(batch.keys())[0]])

            # Keep track of all the slice labels
            all_slice_membership = []

            # Run each slicemaker on the batch
            for slicemaker in slicemakers:
                all_slice_membership.append(
                    slicemaker.apply(
                        slice_membership=np.zeros(
                            (batch_size, slicemaker.num_slices), dtype=np.int32
                        ),
                        batch=batch,
                        columns=columns,
                    )
                )

            # Concatenate all the slice labels
            slice_membership = np.concatenate(all_slice_membership, axis=1)

            # Take the union over the slices (columns)
            slice_membership = np.all(slice_membership, axis=1).astype(np.int32)[
                :, np.newaxis
            ]

            return slice_membership

        return Subpopulation(identifiers=[identifier], apply_fn=apply_fn)


class SubpopulationCollection(Subpopulation):
    def __init__(self, subpopulations: Sequence[Subpopulation], *args, **kwargs):

        super(SubpopulationCollection, self).__init__(
            identifiers=list(
                tz.concat(
                    [subpopulation.identifiers for subpopulation in subpopulations]
                )
            ),
            *args,
            **kwargs,
        )

        # TODO(karan): some subpopulations aren't compatible with each other (e.g.
        #  single column vs. multi column):
        #  add some smarter logic here to handle this

        # Store the subpopulations
        self.subpopulations = subpopulations

    def __call__(
        self,
        batch_or_dataset: Union[Batch, Dataset],
        columns: List[str],
        num_proc: int = None,
        *args,
        **kwargs,
    ):

        if not num_proc or num_proc == 1:
            slices = []
            slice_membership = []
            # Apply each slicebuilder in sequence
            for i, slicebuilder in tqdm(enumerate(self.subpopulations)):
                # Apply the slicebuilder
                slices_i, slice_membership_i = slicebuilder(
                    batch_or_dataset=batch_or_dataset,
                    columns=columns,
                    *args,
                    **kwargs,
                )

                # Add in the slices and slice membership
                slices.extend(slices_i)
                slice_membership.append(slice_membership_i)

        else:
            # TODO(karan): cleanup, make mp.Pool support simpler across the library
            with Pool(num_proc) as pool:
                slices, slice_membership = zip(
                    *pool.map(
                        lambda sb: sb(
                            batch_or_dataset=batch_or_dataset,
                            columns=columns,
                            *args,
                            **kwargs,
                        ),
                        [slicebuilder for slicebuilder in self.subpopulations],
                    )
                )

                # Combine all the slices
                slices = list(tz.concat(slices))

            def _store_updates(batch, indices):

                # Each Subpopulation will generate slices
                for i, subpopulation in enumerate(self.subpopulations):
                    updates = subpopulation.construct_updates(
                        slice_membership=slice_membership[i][indices],
                        columns=columns,
                    )

                    batch = subpopulation.store(
                        batch=batch,
                        updates=updates,
                    )

                return batch

            if isinstance(batch_or_dataset, Dataset):
                batch_or_dataset = batch_or_dataset.map(
                    _store_updates,
                    with_indices=True,
                    batched=True,
                )

                for subpopulation in self.subpopulations:
                    # Update the Dataset's history
                    batch_or_dataset.update_tape(
                        path=[SLICEBUILDERS, subpopulation.category],
                        identifiers=subpopulation.identifiers,
                        columns=columns,
                    )

        # Combine all the slice membership matrices
        slice_membership = np.concatenate(slice_membership, axis=1)

        return slices, slice_membership

    def apply(
        self,
        slice_membership: np.ndarray,
        batch: Batch,
        columns: List[str],
        *args,
        **kwargs,
    ) -> np.ndarray:
        # Each Subpopulation will generate slices
        for subpopulation, end_idx in zip(
            self.subpopulations, np.cumsum([s.num_slices for s in self.subpopulations])
        ):
            # Fill out the slice_membership
            slice_membership[
                :, end_idx - subpopulation.num_slices : end_idx
            ] = subpopulation.apply(
                slice_membership=slice_membership[
                    :, end_idx - subpopulation.num_slices : end_idx
                ],
                batch=batch,
                columns=columns,
            )

        return slice_membership

    # TODO(karan): add combinations for collections



from __future__ import annotations

import copy
import gzip
import logging
import os
import pickle
from collections import defaultdict
from typing import Callable, Dict, List, Mapping, Optional, Union

import cytoolz as tz
import datasets
import numpy as np
import pyarrow as pa
from datasets import DatasetInfo, Features, NamedSplit
from tqdm.auto import tqdm

from robustnessgym.core.dataformats.abstract import AbstractDataset
from robustnessgym.core.tools import convert_to_batch_fn

logger = logging.getLogger(__name__)

Example = Dict
Batch = Dict[str, List]


class InMemoryDataset(AbstractDataset):
    """Class for datasets that are to be stored in memory."""

    def __init__(
        self,
        *args,
        column_names: List[str] = None,
        info: DatasetInfo = None,
        split: Optional[NamedSplit] = None,
    ):

        # Data is a dictionary of lists
        self._data = {}

        # Single argument
        if len(args) == 1:
            assert column_names is None, "Don't pass in column_names."
            # The data is passed in
            data = args[0]

            # `data` is a dictionary
            if isinstance(data, dict) and len(data):
                # Assert all columns are the same length
                self._assert_columns_all_equal_length(data)
                self._data = data

            # `data` is a list
            elif isinstance(data, list) and len(data):
                # Transpose the list of dicts to a dict of lists i.e. a batch
                data = tz.merge_with(list, *data)
                # Assert all columns are the same length
                self._assert_columns_all_equal_length(data)
                self._data = data

            # `data` is a datasets.Dataset
            elif isinstance(data, datasets.Dataset):
                self._data = data[:]
                info, split = data.info, data.split

        # No argument
        elif len(args) == 0:

            # Use column_names to setup the data dictionary
            if column_names:
                self._data = {k: [] for k in column_names}

        # Setup the DatasetInfo
        info = info.copy() if info is not None else DatasetInfo()
        AbstractDataset.__init__(self, info=info, split=split)

        # Create attributes for all columns and visible columns
        self.all_columns = list(self._data.keys())
        self.visible_columns = None

        # Create attributes for visible rows
        self.visible_rows = None

        # Initialization
        self._initialize_state()

        logger.info(
            f"Created `InMemoryDataset` with {len(self)} rows and "
            f"{len(self.column_names)} columns."
        )

    def _set_features(self):
        """Set the features of the dataset."""
        with self.format():
            self.info.features = Features.from_arrow_schema(
                pa.Table.from_pydict(
                    self[:1],
                ).schema
            )

    def _materialize(self):
        # Materialize data, instead of using a reference to an ancestor Dataset
        self._data = {k: self[k] for k in self._data}

        # Reset visible_rows
        self.set_visible_rows(None)

    def add_column(self, column: str, values: List, overwrite=False) -> None:
        """Add a column to the dataset."""

        assert (
            column not in self.all_columns
        ) or overwrite, (
            f"Column `{column}` already exists, set `overwrite=True` to overwrite."
        )
        assert len(values) == len(self), (
            f"`add_column` failed. "
            f"Values length {len(values)} != dataset length {len(self)}."
        )

        if self.visible_rows is not None:
            # Materialize the data
            self._materialize()

        # Add the column
        self._data[column] = list(values)
        self.all_columns.append(column)
        self.visible_columns.append(column)

        # Set features
        self._set_features()

        logger.info(f"Added column `{column}` with length `{len(values)}`.")

    def remove_column(self, column: str) -> None:
        """Remove a column from the dataset."""
        assert column in self.all_columns, f"Column `{column}` does not exist."

        if self.visible_rows is not None:
            # Materialize the data
            self._materialize()

        # Remove the column
        del self._data[column]
        self.all_columns = [col for col in self.all_columns if col != column]
        self.visible_columns = [col for col in self.visible_columns if col != column]

        # Set features
        self._set_features()

        logger.info(f"Removed column `{column}`.")

    def select_columns(self, columns: List[str]) -> Batch:
        """Select a subset of columns."""
        for col in columns:
            assert col in self._data
        return tz.keyfilter(lambda k: k in columns, self._data)

    def _append_to_empty_dataset(self, example_or_batch: Union[Example, Batch]) -> None:
        """Append a batch of data to the dataset when it's empty."""
        # Convert to batch
        batch = self._example_or_batch_to_batch(example_or_batch)

        # TODO(karan): what other data properties need to be in sync here
        self.all_columns = list(batch.keys())
        self.visible_columns = list(batch.keys())

        # Dataset is empty: create the columns and append the batch
        self._data = {k: [] for k in self.column_names}
        for k in self.column_names:
            self._data[k].extend(batch[k])

    def append(
        self,
        example_or_batch: Union[Example, Batch],
    ) -> None:
        """Append a batch of data to the dataset.

        `batch` must have the same columns as the dataset (regardless of
        what columns are visible).
        """
        if not self.column_names:
            return self._append_to_empty_dataset(example_or_batch)

        # Check that example_or_batch has the same format as the dataset
        # TODO(karan): require matching on nested features?
        columns = list(example_or_batch.keys())
        assert set(columns) == set(
            self.column_names
        ), f"Mismatched columns\nbatch: {columns}\ndataset: {self.column_names}"

        # Convert to a batch
        batch = self._example_or_batch_to_batch(example_or_batch)

        # Append to the dataset
        for k in self.column_names:
            self._data[k].extend(batch[k])

    def _remap_index(self, index):
        if isinstance(index, int):
            return self.visible_rows[index].item()
        elif isinstance(index, slice):
            return self.visible_rows[index].tolist()
        elif isinstance(index, str):
            return index
        elif (isinstance(index, tuple) or isinstance(index, list)) and len(index):
            return self.visible_rows[index].tolist()
        elif isinstance(index, np.ndarray) and len(index.shape) == 1:
            return self.visible_rows[index].tolist()
        else:
            raise TypeError("Invalid argument type: {}".format(type(index)))

    def __getitem__(self, index):
        if self.visible_rows is not None:
            # Remap the index if only some rows are visible
            index = self._remap_index(index)

        if (
            isinstance(index, int)
            or isinstance(index, slice)
            or isinstance(index, np.int)
        ):
            # int or slice index => standard list slicing
            return {k: self._data[k][index] for k in self.visible_columns}
        elif isinstance(index, str):
            # str index => column selection
            if index in self.column_names:
                if self.visible_rows is not None:
                    return [self._data[index][i] for i in self.visible_rows]
                return self._data[index]
            raise AttributeError(f"Column {index} does not exist.")
        elif (isinstance(index, tuple) or isinstance(index, list)) and len(index):
            return {k: [self._data[k][i] for i in index] for k in self.visible_columns}
        elif isinstance(index, np.ndarray) and len(index.shape) == 1:
            return {
                k: [self._data[k][int(i)] for i in index] for k in self.visible_columns
            }
        else:
            raise TypeError("Invalid argument type: {}".format(type(index)))

    def update(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        # input_columns: Optional[Union[str, List[str]]] = None,
        batched: bool = False,
        batch_size: Optional[int] = 1000,
        remove_columns: Optional[List[str]] = None,
        **kwargs,
    ) -> Optional[InMemoryDataset]:
        """Update the columns of the dataset."""
        # TODO(karan): make this fn go faster
        # most of the time is spent on the merge, speed it up further

        # Return if the function is None
        if function is None:
            logger.info("`function` None, returning None.")
            return self

        # Return if `self` has no examples
        if not len(self):
            logger.info("Dataset empty, returning None.")
            return self

        # if isinstance(input_columns, str):
        #     input_columns = [input_columns]

        # Set the format
        # if input_columns is not None:
        #     previous_format = self.visible_columns
        #     self.set_format(input_columns)

        # Get some information about the function
        function_properties = self._inspect_function(function, with_indices, batched)
        assert (
            function_properties.dict_output
        ), f"`function` {function} must return dict."

        if not batched:
            # Convert to a batch function
            function = convert_to_batch_fn(function, with_indices=with_indices)
            logger.info(f"Converting `function` {function} to batched function.")

        # Update always returns a new dataset
        logger.info("Running update, a new dataset will be returned.")
        if self.visible_rows is not None:
            # Run .map() to get updated batches and pass them into a new dataset
            new_dataset = InMemoryDataset(
                self.map(
                    (
                        lambda batch, indices: self._merge_batch_and_output(
                            batch, function(batch, indices)
                        )
                    )
                    if with_indices
                    else (
                        lambda batch: self._merge_batch_and_output(
                            batch, function(batch)
                        )
                    ),
                    with_indices=with_indices,
                    batched=True,
                    batch_size=batch_size,
                )
            )
        else:
            if function_properties.updates_existing_column:
                # Copy the ._data dict with a reference to the actual columns
                new_dataset = self.copy()

                # Calculate the values for the updated columns using a .map()
                output = self.map(
                    (
                        lambda batch, indices:
                        # Only merge columns that get updated
                        self._merge_batch_and_output(
                            {
                                k: v
                                for k, v in batch.items()
                                if k in function_properties.existing_columns_updated
                            },
                            function(batch, indices),
                        )
                    )
                    if with_indices
                    else (
                        lambda batch:
                        # Only merge columns that get updated
                        self._merge_batch_and_output(
                            {
                                k: v
                                for k, v in batch.items()
                                if k in function_properties.existing_columns_updated
                            },
                            function(batch),
                        )
                    ),
                    with_indices=with_indices,
                    batched=True,
                    batch_size=batch_size,
                )

                # Add new columns / overwrite existing columns for the update
                for col, vals in output.items():
                    new_dataset.add_column(col, vals, overwrite=True)
            else:
                # Copy the ._data dict with a reference to the actual columns
                new_dataset = self.copy()

                # Calculate the values for the new columns using a .map()
                output = new_dataset.map(
                    function=function,
                    with_indices=with_indices,
                    batched=True,
                    batch_size=batch_size,
                )
                # Add new columns for the update
                for col, vals in output.items():
                    new_dataset.add_column(col, vals)

        # # Update returns a new dataset
        # new_dataset = InMemoryDataset()
        #
        # # Run the update
        # for i, batch in enumerate(self.batch(batch_size)):
        #     # Run the function to compute the update
        #     output = (
        #         function(
        #             batch,
        #             range(i * batch_size, min(len(self), (i + 1) * batch_size)),
        #         )
        #         if with_indices
        #         else function(batch)
        #     )
        #
        #     # Merge the output with the batch
        #     output = self._merge_batch_and_output(batch, output)
        #     new_dataset.append(output)

        # # Update the features of the dataset
        # new_dataset._set_features()

        # Remove columns
        if remove_columns:
            for col in remove_columns:
                new_dataset.remove_column(col)
            logger.info(f"Removed columns {remove_columns}.")
        # Reset the format
        # if input_columns:
        #     self.set_format(previous_format)

        return new_dataset

    def map(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        input_columns: Optional[Union[str, List[str]]] = None,
        batched: bool = False,
        batch_size: Optional[int] = 1000,
        drop_last_batch: bool = False,
        **kwargs,
    ) -> Optional[Dict, List]:
        """Apply a map over the dataset."""

        # Just return if the function is None
        if function is None:
            logger.info("`function` None, returning None.")
            return None

        # Return if `self` has no examples
        if not len(self):
            logger.info("Dataset empty, returning None.")
            return None

        if isinstance(input_columns, str):
            input_columns = [input_columns]

        # Set the format
        previous_format = self.visible_columns
        if input_columns:
            self.set_format(input_columns)

        if not batched:
            # Convert to a batch function
            function = convert_to_batch_fn(function, with_indices=with_indices)
            logger.info(f"Converting `function` {function} to a batched function.")

        # Run the map
        logger.info("Running `map`, the dataset will be left unchanged.")
        outputs = None
        for i, batch in tqdm(
            enumerate(self.batch(batch_size, drop_last_batch)),
            total=(len(self) // batch_size) + (1 - int(drop_last_batch)),
        ):

            # Run `function` on the batch
            output = (
                function(
                    batch,
                    range(i * batch_size, min(len(self), (i + 1) * batch_size)),
                )
                if with_indices
                else function(batch)
            )

            if i == 0:
                # Create an empty dict or list for the outputs
                outputs = defaultdict(list) if isinstance(output, Mapping) else []

            # Append the output
            if output is not None:
                if isinstance(output, Mapping):
                    for k in output.keys():
                        outputs[k].extend(output[k])
                else:
                    outputs.extend(output)

        # Reset the format
        if input_columns:
            self.set_format(previous_format)

        if not len(outputs):
            return None
        elif isinstance(outputs, dict):
            return dict(outputs)
        return outputs

    def filter(
        self,
        function: Optional[Callable] = None,
        with_indices=False,
        input_columns: Optional[Union[str, List[str]]] = None,
        batched: bool = False,
        batch_size: Optional[int] = 1000,
        drop_last_batch: bool = False,
        **kwargs,
    ) -> Optional[InMemoryDataset]:
        """Apply a filter over the dataset."""
        # Just return if the function is None
        if function is None:
            logger.info("`function` None, returning None.")
            return None

        # Return if `self` has no examples
        if not len(self):
            logger.info("Dataset empty, returning None.")
            return None

        # Get some information about the function
        function_properties = self._inspect_function(
            function,
            with_indices,
            batched=batched,
        )
        assert function_properties.bool_output, "function must return boolean."

        # Map to get the boolean outputs and indices
        logger.info("Running `filter`, a new dataset will be returned.")
        outputs = self.map(
            function=function,
            with_indices=with_indices,
            input_columns=input_columns,
            batched=batched,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
        )
        indices = np.where(outputs)[0]

        # Reset the format to set visible columns for the filter
        with self.format():
            # Filter returns a new dataset
            new_dataset = self.copy()
            new_dataset.set_visible_rows(indices)

        return new_dataset

    def copy(self, deepcopy=False):
        """Return a copy of the dataset."""
        if deepcopy:
            return copy.deepcopy(self)
        else:
            dataset = InMemoryDataset()
            dataset.__dict__ = {k: copy.copy(v) for k, v in self.__dict__.items()}
            return dataset

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return {"_data", "all_columns", "visible_rows", "_info", "_split"}

    @classmethod
    def _assert_state_keys(cls, state: Dict) -> None:
        """Assert that a state contains all required keys."""
        assert (
            set(state.keys()) == cls._state_keys()
        ), f"State must contain all state keys: {cls._state_keys()}."

    def __getstate__(self) -> Dict:
        """Get the internal state of the dataset."""
        state = {key: getattr(self, key) for key in self._state_keys()}
        self._assert_state_keys(state)
        return state

    def __setstate__(self, state: Dict) -> None:
        """Set the internal state of the dataset."""
        if not isinstance(state, dict):
            raise ValueError(
                f"`state` must be a dictionary containing " f"{self._state_keys()}."
            )

        self._assert_state_keys(state)

        for key in self._state_keys():
            setattr(self, key, state[key])

        # Do some initialization
        self._initialize_state()

    @classmethod
    def load_from_disk(cls, path: str) -> InMemoryDataset:
        """Load the in-memory dataset from disk."""

        with gzip.open(os.path.join(path, "data.gz")) as f:
            dataset = pickle.load(f)
        # # Empty state dict
        # state = {}
        #
        # # Load the data
        # with gzip.open(os.path.join(path, "data.gz")) as f:
        #     state['_data'] = pickle.load(f)
        #
        # # Load the metadata
        # metadata = json.load(
        #     open(os.path.join(path, "metadata.json"))
        # )
        #
        # # Merge the metadata into the state
        # state = {**state, **metadata}

        # Create an empty `InMemoryDataset` and set its state
        # dataset = cls()
        # dataset.__setstate__(state)

        return dataset

    def save_to_disk(self, path: str):
        """Save the in-memory dataset to disk."""
        # Create all the directories to the path
        os.makedirs(path, exist_ok=True)

        # Store the data in a compressed format
        with gzip.open(os.path.join(path, "data.gz"), "wb") as f:
            pickle.dump(self, f)

        # # Get the dataset state
        # state = self.__getstate__()
        #
        # # Store the data in a compressed format
        # with gzip.open(os.path.join(path, "data.gz"), "wb") as f:
        #     pickle.dump(state['_data'], f)
        #
        # # Store the metadata
        # json.dump(
        #     {k: v for k, v in state.items() if k != '_data'},
        #     open(os.path.join(path, "metadata.json"), 'w'),
        # )



from typing import List, Sequence

import cytoolz as tz
import numpy as np
from datasets import load_metric

from robustnessgym.cachedops.spacy import Spacy
from robustnessgym.core.cachedops import CachedOperation
from robustnessgym.core.dataset import Batch, transpose_batch


class SentenceSimilarityMatrix(CachedOperation):
    def __init__(self):
        super(SentenceSimilarityMatrix, self).__init__()

    def similarity(
        self, batch_sentences_1: List[List[str]], batch_sentences_2: List[List[str]]
    ) -> List:
        raise NotImplementedError("Must implement a similarity computation.")

    def apply(self, batch: Batch, columns: List[str], **kwargs):
        assert len(columns) == 2, "Must specify exactly two columns."

        # Retrieve the sentences in the given columns
        sentences = Spacy.retrieve(
            batch=batch,
            columns=[[col] for col in columns],
            proc_fns=Spacy.sentences,
        )

        return self.similarity(*[sentences[col] for col in columns])


class DocumentSimilarityScore(CachedOperation):
    def __init__(self):
        super(DocumentSimilarityScore, self).__init__()
        self.metric = load_metric("rouge")

    def similarity(self, batch_doc_1: List[str], batch_doc_2: List[str]):
        raise NotImplementedError("Must implement a similarity computation.")

    def apply(self, batch, columns, **kwargs):
        assert len(columns) == 2
        return self.similarity(*[batch[col] for col in columns])


class RougeScore(DocumentSimilarityScore):
    def __init__(self):
        super(RougeScore, self).__init__()
        self.metric = load_metric("rouge")

    def similarity(self, batch_doc_1: List[str], batch_doc_2: List[str]):
        # Compute the scores between every pair of documents
        scores = self.metric.compute(
            predictions=batch_doc_1, references=batch_doc_2, use_agregator=False
        )

        # Transpose the batch of scores
        scores = [
            tz.valmap(
                lambda v: {
                    m: getattr(v, m) for m in ["precision", "recall", "fmeasure"]
                },
                example,
            )
            for example in transpose_batch(scores)
        ]

        return scores

    @classmethod
    def select(
        cls, decoded_batch: List, metric: Sequence[str] = ("rouge1", "fmeasure")
    ):
        if len(metric) == 1:
            return [scores[metric[0]] for scores in decoded_batch]
        elif len(metric) == 2:
            return [scores[metric[0]][metric[1]] for scores in decoded_batch]
        else:
            raise ValueError(f"metric {metric} must be a sequence of length <= 2.")


class RougeMatrix(SentenceSimilarityMatrix):
    def __init__(self):
        super(RougeMatrix, self).__init__()
        self.metric = load_metric("rouge")

    def similarity(
        self, batch_sentences_1: List[List[str]], batch_sentences_2: List[List[str]]
    ):
        batch_similarity = []
        for sents_1, sents_2 in zip(batch_sentences_1, batch_sentences_2):
            # Compute the scores between every pair of sentences
            scores = self.metric.compute(
                predictions=np.repeat(sents_1, len(sents_2)),
                references=sents_2 * len(sents_1),
                use_agregator=False,
            )

            # Organize all the scores into a similarity matrix for each metric
            similarity_mat = {
                k: {
                    m: np.array([getattr(e, m) for e in v])
                    .reshape(len(sents_1), len(sents_2))
                    .tolist()
                    for m in ["precision", "recall", "fmeasure"]
                }
                for k, v in scores.items()
            }

            batch_similarity.append(similarity_mat)

        return batch_similarity

    @classmethod
    def select(
        cls, decoded_batch: List, metric: Sequence[str] = ("rouge1", "fmeasure")
    ):
        if len(metric) == 1:
            return [
                tz.valmap(np.array, matrices[metric[0]]) for matrices in decoded_batch
            ]
        elif len(metric) == 2:
            return [
                np.array(matrices[metric[0]][metric[1]]) for matrices in decoded_batch
            ]
        else:
            raise ValueError(f"metric {metric} must be a sequence of length <= 2.")
